# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import tensorflow as tf
# pylint: disable=no-name-in-module
from tensorflow.python.framework import tensor_shape
from tensorflow.python.keras.layers import Wrapper
from tensorflow.python.layers.convolutional import Conv1D
from tensorflow.python.ops import variable_scope
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.eager import context
from tensorflow.python.ops import nn_impl
from tensorflow.python.keras import initializers
from tensorflow.python.keras.engine.base_layer import InputSpec
from tensorflow.python.ops import array_ops
from tensorflow.python.framework import ops


# ***NOTE***: The WeightNorm Class is copied from this PR:
# https://github.com/tensorflow/tensorflow/issues/14070
# Once this becomes part of the official TF release, it will be removed
class WeightNorm(Wrapper):
    """ This wrapper reparameterizes a layer by decoupling the weight's
    magnitude and direction. This speeds up convergence by improving the
    conditioning of the optimization problem.

    Weight Normalization: A Simple Reparameterization to Accelerate
    Training of Deep Neural Networks: https://arxiv.org/abs/1602.07868
    Tim Salimans, Diederik P. Kingma (2016)

    WeightNorm wrapper works for keras and tf layers.

    ```python
      net = WeightNorm(tf.keras.layers.Conv2D(2, 2, activation='relu'),
             input_shape=(32, 32, 3), data_init=True)(x)
      net = WeightNorm(tf.keras.layers.Conv2D(16, 5, activation='relu'),
                       data_init=True)
      net = WeightNorm(tf.keras.layers.Dense(120, activation='relu'),
                       data_init=True)(net)
      net = WeightNorm(tf.keras.layers.Dense(n_classes),
                       data_init=True)(net)
    ```

    Arguments:
      layer: a layer instance.
      data_init: If `True` use data dependent variable initialization

    Raises:
      ValueError: If not initialized with a `Layer` instance.
      ValueError: If `Layer` does not contain a `kernel` of weights
      NotImplementedError: If `data_init` is True and running graph execution
    """
    def __init__(self, layer, data_init=False, **kwargs):
        if not isinstance(layer, Layer):
            raise ValueError(
                'Please initialize `WeightNorm` layer with a '
                '`Layer` instance. You passed: {input}'.format(input=layer))

        if not context.executing_eagerly() and data_init:
            raise NotImplementedError(
                'Data dependent variable initialization is not available for '
                'graph execution')

        self.initialized = True
        if data_init:
            self.initialized = False

        self.layer_depth = None
        self.norm_axes = None
        super(WeightNorm, self).__init__(layer, **kwargs)
        self._track_checkpointable(layer, name='layer')

    def _compute_weights(self):
        """Generate weights by combining the direction of weight vector
         with it's norm """
        with variable_scope.variable_scope('compute_weights'):
            self.layer.kernel = nn_impl.l2_normalize(
                self.layer.v, axis=self.norm_axes) * self.layer.g

    def _init_norm(self, weights):
        """Set the norm of the weight vector"""
        from tensorflow.python.ops.linalg_ops import norm
        with variable_scope.variable_scope('init_norm'):
            # pylint: disable=no-member
            flat = array_ops.reshape(weights, [-1, self.layer_depth])
            # pylint: disable=no-member
            return array_ops.reshape(norm(flat, axis=0), (self.layer_depth,))

    def _data_dep_init(self, inputs):
        """Data dependent initialization for eager execution"""
        from tensorflow.python.ops.nn import moments
        from tensorflow.python.ops.math_ops import sqrt

        with variable_scope.variable_scope('data_dep_init'):
            # Generate data dependent init values
            activation = self.layer.activation
            self.layer.activation = None
            x_init = self.layer.call(inputs)
            m_init, v_init = moments(x_init, self.norm_axes)
            scale_init = 1. / sqrt(v_init + 1e-10)

        # Assign data dependent init values
        self.layer.g = self.layer.g * scale_init
        self.layer.bias = (-1 * m_init * scale_init)
        self.layer.activation = activation
        self.initialized = True

    # pylint: disable=signature-differs
    def build(self, input_shape):
        """Build `Layer`"""
        input_shape = tensor_shape.TensorShape(input_shape).as_list()
        self.input_spec = InputSpec(shape=input_shape)

        if not self.layer.built:
            self.layer.build(input_shape)
            self.layer.built = False

            if not hasattr(self.layer, 'kernel'):
                raise ValueError(
                    '`WeightNorm` must wrap a layer that'
                    ' contains a `kernel` for weights'
                )

            # The kernel's filter or unit dimension is -1
            self.layer_depth = int(self.layer.kernel.shape[-1])
            self.norm_axes = list(range(self.layer.kernel.shape.ndims - 1))

            self.layer.v = self.layer.kernel
            self.layer.g = self.layer.add_variable(
                name="g",
                shape=(self.layer_depth,),
                initializer=initializers.get('ones'),
                dtype=self.layer.kernel.dtype,
                trainable=True)

            with ops.control_dependencies([self.layer.g.assign(
                    self._init_norm(self.layer.v))]):
                self._compute_weights()

            self.layer.built = True

        super(WeightNorm, self).build()
        self.built = True

    # pylint: disable=arguments-differ
    def call(self, inputs):
        """Call `Layer`"""
        if context.executing_eagerly():
            if not self.initialized:
                self._data_dep_init(inputs)
            self._compute_weights()  # Recompute weights for each forward pass

        output = self.layer.call(inputs)
        return output

    def compute_output_shape(self, input_shape):
        return tensor_shape.TensorShape(
            self.layer.compute_output_shape(input_shape).as_list())


class TCN:
    """
    This class defines core TCN architecture.
    This is only the base class, training strategy is not implemented.
    """
    def __init__(self, max_len, n_features_in, hidden_sizes, kernel_size=7, dropout=0.2):
        """
        To use this class,
            1. Inherit this class
            2. Define the training losses in build_train_graph()
            3. Define the training strategy in run()
            4. After the inherited class object is initialized,
               call build_train_graph followed by run

        Args:
            max_len: Maximum length of sequence
            n_features_in: Number of input features (dimensions)
            hidden_sizes: Number of hidden sizes in each layer of TCN (same for all layers)
            kernel_size: Kernel size of convolution filter (same for all layers)
            dropout: Dropout, fraction of activations to drop
        """
        self.max_len = max_len
        self.n_features_in = n_features_in
        self.hidden_sizes = hidden_sizes
        self.kernel_size = kernel_size
        self.dropout = dropout
        self.n_hidden_layers = len(self.hidden_sizes)
        receptive_field_len = self.calculate_receptive_field()
        if receptive_field_len < self.max_len:
            print("Warning! receptive field of the TCN: "
                  "%d is less than the input sequence length: %d."
                  % (receptive_field_len, self.max_len))
        else:
            print("Receptive field of the TCN: %d, input sequence length: %d."
                  % (receptive_field_len, self.max_len))
        self.layer_activations = []

        # toggle this for train/inference mode
        self.training_mode = tf.placeholder(tf.bool, name='training_mode')

        self.sequence_output = None

    def calculate_receptive_field(self):
        """

        Returns:

        """
        return 1 + 2 * (self.kernel_size - 1) * (2 ** self.n_hidden_layers - 1)

    def build_network_graph(self, x, last_timepoint=False):
        """
        Given the input placeholder x, build the entire TCN graph
        Args:
            x: Input placeholder
            last_timepoint: Whether or not to select only the last timepoint to output

        Returns:
            output of the TCN
        """
        # loop and define multiple residual blocks
        with tf.variable_scope("tcn"):
            for i in range(self.n_hidden_layers):
                dilation_size = 2 ** i
                in_channels = self.n_features_in if i == 0 else self.hidden_sizes[i - 1]
                out_channels = self.hidden_sizes[i]
                with tf.variable_scope("residual_block_" + str(i)):
                    x = self._residual_block(x, in_channels, out_channels, dilation_size,
                                             (self.kernel_size - 1) * dilation_size)
                    x = tf.nn.relu(x)
                self.layer_activations.append(x)
            self.sequence_output = x

            # get outputs
            if not last_timepoint:
                prediction = self.sequence_output
            else:
                # last time point size (batch_size, hidden_sizes_encoder)
                width = self.sequence_output.shape[1].value
                lt = tf.squeeze(tf.slice(self.sequence_output, [0, width - 1, 0],
                                         [-1, 1, -1]), axis=1)
                prediction = \
                    tf.layers.Dense(1, kernel_initializer=tf.initializers.random_normal(0, 0.01),
                                    bias_initializer=tf.initializers.random_normal(0, 0.01))(lt)

        return prediction

    def _residual_block(self, x, in_channels, out_channels, dilation, padding):
        """
        Defines the residual block
        Args:
            x: Input tensor to residual block
            in_channels: Number of input features (dimensions)
            out_channels: Number of output features (dimensions)
            dilation: Dilation rate
            padding: Padding value

        Returns:
            Output of residual path
        """
        xin = x
        # define two temporal blocks
        for i in range(2):
            with tf.variable_scope("temporal_block_" + str(i)):
                x = self._temporal_block(x, out_channels, dilation, padding)

        # sidepath
        if in_channels != out_channels:
            x_side = tf.layers.Conv1D(filters=out_channels, kernel_size=1, padding='same',
                                      strides=1, activation=None, dilation_rate=1,
                                      kernel_initializer=tf.initializers.random_normal(0, 0.01),
                                      bias_initializer=tf.initializers.random_normal(0, 0.01))(xin)
        else:
            x_side = xin

        # combine both
        return tf.add(x, x_side)

    def _temporal_block(self, x, out_channels, dilation, padding):
        """
        Defines the temporal block, which is a dilated causual conv layer,
        followed by relu and dropout
        Args:
            x: Input to temporal block
            out_channels: Number of conv filters
            dilation: dilation rate
            padding: padding value

        Returns:
            Tensor output of temporal block
        """
        # conv layer
        x = self._dilated_causal_conv(x, out_channels, dilation, padding)

        x = tf.nn.relu(x)

        # dropout
        batch_size = tf.shape(x)[0]
        x = tf.layers.dropout(x, rate=self.dropout, noise_shape=[batch_size, 1, out_channels],
                              training=self.training_mode)

        return x

    # define model
    def _dilated_causal_conv(self, x, n_filters, dilation, padding):
        """
        Defines dilated causal convolution
        Args:
            x: Input activation
            n_filters: Number of convolution filters
            dilation: Dilation rate
            padding: padding value

        Returns:
            Tensor output of convolution
        """
        input_width = x.shape[1].value
        with tf.variable_scope("dilated_causal_conv"):
            # define dilated convolution layer with left side padding
            x = tf.pad(x, tf.constant([[0, 0], [padding, 0], [0, 0]]), 'CONSTANT')
            x = WeightNorm(Conv1D(filters=n_filters, kernel_size=self.kernel_size, padding='valid',
                                  strides=1, activation=None, dilation_rate=dilation,
                                  kernel_initializer=tf.initializers.random_normal(0, 0.01),
                                  bias_initializer=tf.initializers.random_normal(0, 0.01)))(x)

        assert x.shape[1].value == input_width

        return x

    def build_train_graph(self, *args, **kwargs):
        """
        Placeholder for defining training losses and metrics
        """
        raise NotImplementedError("Error! losses for training must be defined")

    def run(self, *args, **kwargs):
        """
        Placeholder for defining training strategy
        """
        raise NotImplementedError("Error! training routine must be defined")


class CommonLayers:
    """
    Class that contains the common layers for language modeling -
            word embeddings and projection layer
    """
    def __init__(self):
        """
        Initialize class
        """
        self.word_embeddings_tf = None
        self.num_words = None
        self.n_features_in = None

    def define_input_layer(self, input_placeholder_tokens, word_embeddings,
                           embeddings_trainable=True):
        """
        Define the input word embedding layer
        Args:
            input_placeholder_tokens: tf.placeholder, input to the model
            word_embeddings: numpy array (optional), to initialize the embeddings with
            embeddings_trainable: boolean, whether or not to train the embedding table

        Returns:
            Embeddings corresponding to the data in input placeholder
        """
        with tf.device('/cpu:0'):
            with tf.variable_scope("embedding_layer", reuse=False):
                if word_embeddings is None:
                    initializer = tf.initializers.random_normal(0, 0.01)
                else:
                    initializer = tf.constant_initializer(word_embeddings)
                self.word_embeddings_tf = tf.get_variable("embedding_table",
                                                          shape=[self.num_words,
                                                                 self.n_features_in],
                                                          initializer=initializer,
                                                          trainable=embeddings_trainable)

                input_embeddings = tf.nn.embedding_lookup(self.word_embeddings_tf,
                                                          input_placeholder_tokens)
        return input_embeddings

    def define_projection_layer(self, prediction, tied_weights=True):
        """
        Define the output word embedding layer
        Args:
            prediction: tf.tensor, the prediction from the model
            tied_weights: boolean, whether or not to tie weights from the input embedding layer

        Returns:
            Probability distribution over vocabulary
        """
        with tf.device('/cpu:0'):
            if tied_weights:
                # tie projection layer and embedding layer
                with tf.variable_scope("embedding_layer", reuse=tf.AUTO_REUSE):
                    softmax_w = tf.matrix_transpose(self.word_embeddings_tf)
                    softmax_b = tf.get_variable('softmax_b', [self.num_words])
                    _, l, k = prediction.shape.as_list()
                    prediction_reshaped = tf.reshape(prediction, [-1, k])
                    mult_out = tf.nn.bias_add(tf.matmul(prediction_reshaped, softmax_w), softmax_b)
                    projection_out = tf.reshape(mult_out, [-1, l, self.num_words])
            else:
                with tf.variable_scope("projection_layer", reuse=False):
                    projection_out = tf.layers.Dense(self.num_words)(prediction)
        return projection_out



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import argparse
import logging
import pickle

import numpy as np

from nlp_architect.models.cross_doc_coref.system.cdc_utils import load_mentions_vocab_from_files
from nlp_architect.utils import io

logging.basicConfig(level=logging.DEBUG)
logger = logging.getLogger(__name__)

parser = argparse.ArgumentParser(description='Create GloVe dataset only dump')

parser.add_argument('--glove', type=str, help='glove db file', required=True)

parser.add_argument('--mentions', type=str, help='mentions file', required=True)

parser.add_argument('--output', type=str, help='location were to create dump file', required=True)

args = parser.parse_args()


def load_glove_for_vocab(glove_filename, vocabulary):
    vocab = []
    embd = []
    with open(glove_filename) as glove_file:
        for line in glove_file:
            row = line.strip().split(' ')
            word = row[0]
            if word in vocabulary:
                vocab.append(word)
                embd.append(row[1:])
    logger.info('Loaded GloVe!')

    embeddings = np.asarray(embd, dtype=float)
    word_to_ix = {word: i for i, word in enumerate(vocab)}
    return word_to_ix, embeddings


def glove_dump():
    filter_stop_words = False
    glove_file = args.glove
    out_file = args.output
    mention_files = [args.mentions]
    vocab = load_mentions_vocab_from_files(mention_files, filter_stop_words)
    word_to_ix, embeddings = load_glove_for_vocab(glove_file, vocab)

    logger.info('Words in vocabulary %d', len(vocab))
    logger.info('Found %d words from vocabulary', len(word_to_ix.keys()))
    with open(out_file, 'wb') as f:
        pickle.dump([word_to_ix, embeddings], f)
    logger.info('Saving dump to file-%s', out_file)


if __name__ == '__main__':
    io.validate_existing_filepath(args.mentions)
    io.validate_existing_filepath(args.output)
    io.validate_existing_filepath(args.glove)
    glove_dump()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import time
from pathlib import Path
from os import PathLike
from nlp_architect.models.absa import TRAIN_OUT
from nlp_architect.models.absa.train.acquire_terms import AcquireTerms
from nlp_architect.models.absa.train.rerank_terms import RerankTerms
from nlp_architect.models.absa.utils import parse_docs, _download_pretrained_rerank_model
from nlp_architect.utils.io import download_unzip

EMBEDDING_URL = 'http://nlp.stanford.edu/data', 'glove.840B.300d.zip'
EMBEDDING_PATH = TRAIN_OUT / 'word_emb_unzipped' / 'glove.840B.300d.txt'
RERANK_MODEL_DEFAULT_PATH = rerank_model_dir = TRAIN_OUT / 'reranking_model' / 'rerank_model.h5'


class TrainSentiment(object):
    def __init__(self, parse: bool = True, rerank_model: PathLike = None):
        self.start_time = time.time()
        self.acquire_lexicon = AcquireTerms()
        if parse:
            from nlp_architect.pipelines.spacy_bist import SpacyBISTParser
            self.parser = SpacyBISTParser()
        else:
            self.parser = None

        if not rerank_model:
            print('using pre-trained reranking model')
            rerank_model = _download_pretrained_rerank_model(RERANK_MODEL_DEFAULT_PATH)

        download_unzip(*EMBEDDING_URL, EMBEDDING_PATH, license_msg="Glove word embeddings.")
        self.rerank = RerankTerms(vector_cache=True, rerank_model=rerank_model,
                                  emb_model_path=EMBEDDING_PATH)

    def run(self, data: PathLike = None, parsed_data: PathLike = None):
        if not parsed_data:
            if not self.parser:
                raise RuntimeError("Parser not initialized (try parse=True at init )")
            parsed_dir = TRAIN_OUT / 'parsed' / Path(data).stem
            self.parse_data(data, parsed_dir)
            parsed_data = parsed_dir

        generated_aspect_lex = self.acquire_lexicon.acquire_lexicons(parsed_data)
        generated_opinion_lex_reranked = \
            self.rerank.predict(AcquireTerms.acquired_opinion_terms_path,
                                AcquireTerms.generic_opinion_lex_path)
        return generated_opinion_lex_reranked, generated_aspect_lex

    def parse_data(self, data: PathLike, parsed_dir: PathLike):
        _, data_size = parse_docs(self.parser, data, out_dir=parsed_dir)
        if data_size < 1000:
            raise ValueError('The data contains only {0} sentences. A minimum of 1000 '
                             'sentences is required for training.'.format(data_size))



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

"""
Script that prepares the input corpus for np2vec training: it runs NP extractor on the corpus and
marks extracted NP's.
"""
import gzip
import json
import logging
import sys
from argparse import ArgumentParser
from os import path, makedirs

from tqdm import tqdm

from nlp_architect import LIBRARY_OUT
from nlp_architect.pipelines.spacy_np_annotator import NPAnnotator, get_noun_phrases
from nlp_architect.utils.io import check_size, download_unlicensed_file, validate_parent_exists
from nlp_architect.utils.text import spacy_normalizer, SpacyInstance

logging.basicConfig(stream=sys.stdout, level=logging.INFO)
logger = logging.getLogger(__name__)

np2id = {}
id2group = {}
id2rep = {}
np2count = {}
nlp_chunker_url = 'http://nervana-modelzoo.s3.amazonaws.com/NLP/chunker/'
chunker_path = str(LIBRARY_OUT / 'chunker-pretrained')
chunker_model_dat_file = 'model_info.dat.params'
chunker_model_file = 'model.h5'


def get_group_norm(spacy_span):
    """
    Give a span, determine the its group and return the normalized text representing the group

    Args:
            spacy_span (spacy.tokens.Span)
    """
    np = spacy_span.text
    norm = spacy_normalizer(np, spacy_span.lemma_)
    if args.mark_char in norm:
        norm = norm.replace(args.mark_char, ' ')
    if np not in np2count:  # new np
        np2count[np] = 1
        np2id[np] = norm
        if norm in id2group:  # norm already exist
            id2group[norm].append(np)
        else:
            id2group[norm] = [np]
            id2rep[norm] = np
    else:  # another occurrence of this np. norm must exist and be consistent
        np2count[np] += 1
        if np2id[np] != norm:  # new norm to the same np - merge groups.
            #  no need to update np2id[np]
            norm = merge_groups(np, np2id[np], norm)  # set to the already exist
            #  norm so I know this norm is already in id2group/id2rep
        else:  # update rep
            if np2count[np] > np2count[id2rep[norm]]:
                id2rep[norm] = np  # replace rep

    return norm


def load_parser(chunker):
    # load spacy parser
    logger.info('loading spacy. chunker=%s', chunker)
    if 'nlp_arch' in chunker:
        parser = SpacyInstance(model='en_core_web_sm',
                               disable=['textcat', 'ner', 'parser']).parser
        parser.add_pipe(parser.create_pipe('sentencizer'), first=True)
        _path_to_model = path.join(chunker_path, chunker_model_file)
        _path_to_params = path.join(chunker_path, chunker_model_dat_file)
        if not path.exists(chunker_path):
            makedirs(chunker_path)
        if not path.exists(_path_to_model):
            logger.info(
                'The pre-trained model to be downloaded for NLP Architect'
                ' word chunker model is licensed under Apache 2.0')
            download_unlicensed_file(nlp_chunker_url, chunker_model_file, _path_to_model)
        if not path.exists(_path_to_params):
            download_unlicensed_file(nlp_chunker_url, chunker_model_dat_file, _path_to_params)
        parser.add_pipe(NPAnnotator.load(_path_to_model, _path_to_params),
                        last=True)
    else:
        parser = SpacyInstance(model='en_core_web_sm', disable=['textcat', 'ner']).parser
    logger.info('spacy loaded')
    return parser


def extract_noun_phrases(docs, nlp_parser, chunker):
    logger.info('extract nps from: %s', docs)
    spans = []
    for doc in nlp_parser.pipe(docs, n_threads=-1):
        if 'nlp_arch' in chunker:
            spans.extend(get_noun_phrases(doc))
        else:
            nps = list(doc.noun_chunks)
            spans.extend(nps)
    logger.info('nps= %s', str(spans))
    return spans


# pylint: disable-msg=too-many-nested-blocks,too-many-branches
def mark_noun_phrases(corpus_file, marked_corpus_file, nlp_parser, lines_count, chunker,
                      mark_char='_', grouping=False):
    i = 0
    with tqdm(total=lines_count) as pbar:
        for doc in nlp_parser.pipe(corpus_file, n_threads=-1):
            if 'nlp_arch' in chunker:
                spans = get_noun_phrases(doc)
            else:
                spans = list(doc.noun_chunks)
            i += 1
            if len(spans) > 0:
                span = spans.pop(0)
            else:
                span = None
            span_written = False
            for token in doc:
                if span is None:
                    if len(token.text.strip()) > 0:
                        marked_corpus_file.write(token.text + ' ')
                else:
                    if token.idx < span.start_char or token.idx >= span.end_char:  # outside a
                        # span
                        if len(token.text.strip()) > 0:
                            marked_corpus_file.write(token.text + ' ')
                    else:
                        if not span_written:
                            # mark NP's
                            if len(span.text) > 1 and span.lemma_ != '-PRON-':
                                if grouping:
                                    text = get_group_norm(span)
                                else:
                                    text = span.text
                                # mark NP's
                                text = text.replace(' ',
                                                    mark_char) + mark_char
                                marked_corpus_file.write(text + ' ')
                            else:
                                marked_corpus_file.write(span.text + ' ')
                            span_written = True
                        if token.idx + len(token.text) == span.end_char:
                            if len(spans) > 0:
                                span = spans.pop(0)
                            else:
                                span = None
                            span_written = False
            marked_corpus_file.write('\n')
            pbar.update(1)


def merge_groups(np, old_id, diff_id):
    if diff_id in id2group:
        for term in id2group[diff_id]:  # for each term update dicts
            np2id[term] = old_id
            id2group[old_id].append(term)
            if np2count[term] > np2count[id2rep[old_id]]:
                id2rep[old_id] = term
        id2rep.pop(diff_id)
        id2group.pop(diff_id)
    else:
        if np2count[np] > np2count[id2rep[old_id]]:
            id2rep[old_id] = np
    return old_id


if __name__ == '__main__':
    arg_parser = ArgumentParser(__doc__)
    arg_parser.add_argument(
        '--corpus',
        help='path to the input corpus. Compressed files (gz) are also supported. By default, '
             'it is a subset of English Wikipedia. '
             'get subset of English wikipedia from '
             'https://github.com/NervanaSystems/nlp-architect/raw/'
             'master/datasets/wikipedia/enwiki-20171201_subset.txt.gz')
    arg_parser.add_argument(
        '--marked_corpus',
        default='enwiki-20171201_subset_marked.txt',
        type=validate_parent_exists,
        help='path to the marked corpus corpus.')
    arg_parser.add_argument(
        '--mark_char',
        default='_',
        type=str,
        action=check_size(1, 2),
        help='special character that marks NP\'s in the corpus (word separator and NP suffix). '
             'Default value is _.')
    arg_parser.add_argument(
        '--grouping',
        action='store_true',
        default=False,
        help='perform noun-phrase grouping')
    arg_parser.add_argument(
        '--chunker', type=str,
        choices=['spacy', 'nlp_arch'],
        default='spacy',
        help='chunker to use for detecting noun phrases. \'spacy\' for using spacy built-in '
             'chunker or \'nlp_arch\' for NLP Architect NP Extractor')

    args = arg_parser.parse_args()
    if args.corpus.endswith('gz'):
        open_func = gzip.open
        mode = 'rt'
    else:
        open_func = open
        mode = 'r'

    with open_func(args.corpus, mode, encoding='utf8', errors='ignore') as my_corpus_file:
        with open(args.marked_corpus, 'w', encoding='utf8') as my_marked_corpus_file:
            nlp = load_parser(args.chunker)
            num_lines = sum(1 for line in my_corpus_file)
            my_corpus_file.seek(0)
            logger.info('%i lines in corpus', num_lines)
            mark_noun_phrases(my_corpus_file, my_marked_corpus_file, nlp, num_lines,
                              mark_char=args.mark_char, grouping=args.grouping,
                              chunker=args.chunker)

        # write grouping data :
        if args.grouping:
            corpus_dir = path.dirname(args.marked_corpus)
            with open(path.join(corpus_dir, 'id2group'), 'w', encoding='utf8') as id2group_file:
                id2group_file.write(json.dumps(id2group))

            with open(path.join(corpus_dir, 'id2rep'), 'w', encoding='utf8') as id2rep_file:
                id2rep_file.write(json.dumps(id2rep))

            with open(path.join(corpus_dir, 'np2id'), 'w', encoding='utf8') as np2id_file:
                np2id_file.write(json.dumps(np2id))



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

# This dataset should be downloaded from http://jmcauley.ucsd.edu/data/amazon/
# The terms and conditions of the data set license apply.
# Intel does not grant any rights to the data files.
# The Amazon Review Dataset was published in the following papers:
#
# Ups and downs:
# Modeling the visual evolution of fashion trends with one-class collaborative filtering
# R. He, J. McAuley
# WWW, 2016
# http://cseweb.ucsd.edu/~jmcauley/pdfs/www16a.pdf
#
# Image-based recommendations on styles and substitutes
# J. McAuley, C. Targett, J. Shi, A. van den Hengel
# SIGIR, 2015
# http://cseweb.ucsd.edu/~jmcauley/pdfs/sigir15.pdf

import pandas as pd
import json

from nlp_architect.utils.generic import normalize, balance


good_columns = [
    "overall",
    "reviewText",
    "summary"
]


def review_to_sentiment(review):
    # Review is coming in as overall (the rating, reviewText, and summary)
    # this then cleans the summary and review and gives it a positive or negative value
    norm_text = normalize(review[2] + " " + review[1])
    review_sent = ['neutral', norm_text]
    if review[0] > 3:
        review_sent = ['positive', norm_text]
    elif review[0] < 3:
        review_sent = ['negative', norm_text]

    return review_sent


class Amazon_Reviews(object):
    """
    Take the *.json file of Amazon reviews as downloaded from
    http://jmcauley.ucsd.edu/data/amazon/
    Then does data cleaning and balancing, as well as transforms the reviews 1-5 to a sentiment
    """
    def __init__(self, review_file, run_balance=True):
        self.run_balance = run_balance

        print("Parsing and processing json file")
        data = []

        with open(review_file, 'r') as f:
            for line in f:
                data_line = json.loads(line)
                selected_row = []
                for item in good_columns:
                    selected_row.append(data_line[item])
                # as we read in, clean
                data.append(review_to_sentiment(selected_row))

        # Not sure how to easily balance outside of pandas...but should replace eventually
        self.amazon = pd.DataFrame(data, columns=['Sentiment', 'clean_text'])
        self.all_text = self.amazon['clean_text']
        self.labels_0 = pd.get_dummies(self.amazon['Sentiment'])
        self.labels = self.labels_0.values
        self.text = self.amazon['clean_text'].values

    def process(self):
        self.amazon = self.amazon[self.amazon['Sentiment'].isin(['positive', 'negative'])]

        if self.run_balance:
            # balance it out
            self.amazon = balance(self.amazon)

        print("Sample Data")
        print(self.amazon[['Sentiment', 'clean_text']].head())

        # mapping of the labels with dummies (has headers)
        self.labels_0 = pd.get_dummies(self.amazon['Sentiment'])
        self.labels = self.labels_0.values
        self.text = self.amazon['clean_text'].values



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import logging
import pickle
from typing import List

import numpy as np

from nlp_architect.common.cdc.mention_data import MentionDataLight
from nlp_architect.utils.embedding import ELMoEmbedderTFHUB

logger = logging.getLogger(__name__)


class ElmoEmbedding(object):
    def __init__(self):
        logger.info('Loading Elmo Embedding module')
        self.embeder = ELMoEmbedderTFHUB()
        self.cache = dict()
        logger.info('Elmo Embedding module lead successfully')

    def get_head_feature_vector(self, mention: MentionDataLight):
        if mention.mention_context is not None and mention.mention_context:
            sentence = ' '.join(mention.mention_context)
            return self.apply_get_from_cache(sentence, True, mention.tokens_number)

        sentence = mention.tokens_str
        return self.apply_get_from_cache(sentence, False, [])

    def apply_get_from_cache(self, sentence: str, context: bool = False, indexs: List[int] = None):
        if context and indexs is not None:
            if sentence in self.cache:
                elmo_full_vec = self.cache[sentence]
            else:
                elmo_full_vec = self.embeder.get_vector(sentence.split())
                self.cache[sentence] = elmo_full_vec

            elmo_ret_vec = self.get_mention_vec_from_sent(elmo_full_vec, indexs)
        else:
            if sentence in self.cache:
                elmo_ret_vec = self.cache[sentence]
            else:
                elmo_ret_vec = self.get_elmo_avg(sentence.split())
                self.cache[sentence] = elmo_ret_vec

        return elmo_ret_vec

    def get_avrg_feature_vector(self, tokens_str):
        if tokens_str is not None:
            return self.apply_get_from_cache(tokens_str)
        return None

    def get_elmo_avg(self, sentence):
        sentence_embedding = self.embeder.get_vector(sentence)
        return np.mean(sentence_embedding, axis=0)

    @staticmethod
    def get_mention_vec_from_sent(sent_vec, indexs):
        if len(indexs) > 1:
            elmo_ret_vec = np.mean(sent_vec[indexs[0]: indexs[-1] + 1], axis=0)
        else:
            elmo_ret_vec = sent_vec[indexs[0]]

        return elmo_ret_vec


class ElmoEmbeddingOffline(object):
    def __init__(self, dump_file):
        logger.info('Loading Elmo Offline Embedding module')

        if dump_file is not None:
            with open(dump_file, 'rb') as out:
                self.embeder = pickle.load(out)
        else:
            logger.warning('Elmo Offline without loaded embeder!')

        logger.info('Elmo Offline Embedding module lead successfully')

    def get_head_feature_vector(self, mention: MentionDataLight):
        embed = None
        if mention.mention_context is not None and mention.mention_context:
            sentence = ' '.join(mention.mention_context)
            if sentence in self.embeder:
                elmo_full_vec = self.embeder[sentence]
                return ElmoEmbedding.get_mention_vec_from_sent(
                    elmo_full_vec, mention.tokens_number)

        sentence = mention.tokens_str
        if sentence in self.embeder:
            embed = self.embeder[sentence]

        return embed

    def get_avrg_feature_vector(self, tokens_str):
        embed = None
        if tokens_str in self.embeder:
            embed = self.embeder[tokens_str]

        return embed



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import csv
import re
from abc import abstractmethod


class Anonymiser(object):
    """Abstract class for anonymiser algorithm, intended for privacy keeping."""
    @abstractmethod
    def run(self, text):
        pass


class TweetAnonymiser(Anonymiser):
    """Anonymiser for tweets which uses lexicon for simple string replacements."""
    def __init__(self, lexicon_path):
        self.entity_dict = self._init_entity_dict(lexicon_path)

    @staticmethod
    def _init_entity_dict(lexicon_path):
        ret = {}
        with open(lexicon_path) as f:
            for row in csv.reader(f):
                ret[row[0]] = [_ for _ in row[1:] if _]
        return ret

    def run(self, text):
        for anonymised, entities in self.entity_dict.items():
            for entity in entities:
                text = re.sub(entity, anonymised, text, flags=re.IGNORECASE)
        text = ' '.join(["@other_entity" if (word.startswith('@') and word[1:]
                                             not in self.entity_dict.keys())
                         else word for word in text.split()])
        return text



#!/usr/bin/env python
# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import io
import os
import platform
import subprocess
import sys

from setuptools import setup, find_packages

root = os.path.abspath(os.path.dirname(__file__))


# required packages for NLP Architect
with open('requirements.txt') as fp:
    install_requirements = fp.readlines()

# check if GPU available
p = subprocess.Popen(['command -v nvidia-smi'], stdout=subprocess.PIPE, shell=True)
out = p.communicate()[0].decode('utf8')
gpu_available = len(out) > 0

# check python version
py3_ver = int(platform.python_version().split('.')[1])

# Tensorflow version (make sure CPU/MKL/GPU versions exist before changing)
tf_version = '1.12.0'
tf_mkl_url = 'https://storage.googleapis.com/intel-optimized-tensorflow/tensorflow-{}-cp3{}-cp3{}m-linux_x86_64.whl'

# default TF is CPU
chosen_tf = 'tensorflow=={}'.format(tf_version)
# check system is linux for MKL/GPU backends
if 'linux' in sys.platform:
    system_type = 'linux'
    tf_be = os.getenv('NLP_ARCHITECT_BE', False)
    if tf_be and 'mkl' == tf_be.lower():
        if py3_ver == 5 or py3_ver == 6:
            tf_mkl_url_real = tf_mkl_url.format(tf_version, py3_ver, py3_ver)
            subprocess.call([sys.executable, '-m', 'pip', 'install', tf_mkl_url_real])
    elif tf_be and 'gpu' == tf_be.lower() and gpu_available:
        chosen_tf = 'tensorflow-gpu=={}'.format(tf_version)

for r in install_requirements:
    if r.startswith('tensorflow=='):
        install_requirements[install_requirements.index(r)] = chosen_tf

with open('README.md', encoding='utf8') as fp:
    long_desc = fp.read()

with io.open(os.path.join(root, 'nlp_architect', 'version.py'), encoding='utf8') as f:
    version_f = {}
    exec(f.read(), version_f)
    version = version_f['NLP_ARCHITECT_VERSION']

setup(name='nlp-architect',
      version=version,
      description='Intel AI Lab\'s open-source NLP and NLU research library',
      long_description=long_desc,
      long_description_content_type='text/markdown',
      keywords='NLP NLU deep learning natural language processing tensorflow keras dynet',
      author='Intel AI Lab',
      author_email='nlp_architect@intel.com',
      url='https://github.com/NervanaSystems/nlp-architect',
      license='Apache 2.0',
      python_requires='>=3.6.*',
      packages=find_packages(exclude=['tests.*', 'tests', '*.tests', '*.tests.*',
                                      'examples.*', 'examples', '*.examples', '*.examples.*']),
      install_requires=install_requirements,
      scripts=['nlp_architect/nlp_architect'],
      include_package_data=True,
      classifiers=[
          'Development Status :: 3 - Alpha',
          'Intended Audience :: End Users/Desktop',
          'Intended Audience :: Developers',
          'Intended Audience :: Science/Research',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.6',
          'Topic :: Scientific/Engineering',
          'Topic :: Scientific/Engineering :: ' +
          'Artificial Intelligence',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: ' +
          'Python Modules',
          'Topic :: Scientific/Engineering :: Information Analysis',
          'Environment :: Console',
          'Environment :: Web Environment',
          'Operating System :: POSIX',
          'Operating System :: MacOS :: MacOS X',
      ]
      )



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import csv
import json
from os import walk, path, makedirs, PathLike, listdir
from os.path import join, isfile, isdir
from pathlib import Path

import numpy as np

from nlp_architect.common.core_nlp_doc import CoreNLPDoc
from nlp_architect.models.absa import INFERENCE_LEXICONS, TRAIN_LEXICONS
from nlp_architect.models.absa.inference.data_types import LexiconElement, Polarity
from nlp_architect.models.absa.train.data_types import OpinionTerm, string_list_headers
from nlp_architect.utils.io import download_unlicensed_file


def _download_pretrained_rerank_model(rerank_model_full_path):
    rerank_model_dir = path.dirname(rerank_model_full_path)
    if not path.isfile(rerank_model_full_path):
        makedirs(rerank_model_dir, exist_ok=True)
        print("dowloading pre-trained reranking model..")
        download_unlicensed_file('https://s3-us-west-2.amazonaws.com/nlp-architect-data/models/'
                                 'absa/', 'rerank_model.h5', rerank_model_full_path)
    return rerank_model_full_path


def _walk_directory(directory):
    """Iterates a directory's text files and their contents."""
    for dir_path, _, filenames in walk(directory):
        for filename in filenames:
            file_path = join(dir_path, filename)
            if isfile(file_path) and not filename.startswith('.'):
                with open(file_path, 'r', encoding='utf-8') as file:
                    print('Reading ' + filename)
                    doc_text = file.read()
                    yield filename, doc_text


def parse_docs(parser, docs, out_dir=None, show_tok=True, show_doc=True):
    """Parse raw documents in the form of text files in a directory or lines in a text file.

    Args:
        parser (SpacyBISTParser)
        docs (str or PosixPath)
        out_dir (PosixPath): If specified, the output will also be written to this path.
        show_tok (bool, optional): Specifies whether to include token text in output.
        show_doc (bool, optional): Specifies whether to include document text in output.

    Returns:
        (list of CoreNLPDoc)
    """
    if out_dir:
        out_dir.mkdir(parents=True, exist_ok=True)

    params = parser, Path(docs), out_dir, show_tok, show_doc
    parsed_docs = list(parse_dir(*params) if isdir(docs) else parse_txt(*params))
    total_parsed = np.sum([len(doc) for doc in parsed_docs])
    return parsed_docs, total_parsed


def parse_txt(parser, txt_path, out_dir=None, show_tok=True, show_doc=True):
    """Parse raw documents in the form of lines in a text file.

    Args:
        parser (SpacyBISTParser)
        txt_path (PosixPath)
        out_dir (PosixPath): If specified, the output will also be written to this path.
        show_tok (bool, optional): Specifies whether to include token text in output.
        show_doc (bool, optional): Specifies whether to include document text in output.

    Yields:
        CoreNLPDoc: the annotated document.
    """
    with open(txt_path, encoding='utf-8') as f:
        for i, doc_text in enumerate(f):
            print('Parsing document no. {}'.format(i + 1))
            parsed_doc = parser.parse(doc_text.rstrip('\n'), show_tok, show_doc)

            if out_dir:
                out_path = out_dir / (str(i + 1) + '.json')
                print('Dumping parsed document to {}'.format(out_path))
                with open(out_path, 'w', encoding='utf-8') as doc_file:
                    doc_file.write(parsed_doc.pretty_json())

            yield parsed_doc


def parse_dir(parser, input_dir, out_dir=None, show_tok=True, show_doc=True):
    """Parse a directory of raw text documents, one by one.

    Args:
        parser (SpacyBISTParser)
        input_dir (PosixPath)
        out_dir (PosixPath): If specified, the output will also be written to this path.
        show_tok (bool, optional): Specifies whether to include token text in output.
        show_doc (bool, optional): Specifies whether to include document text in output.

    Yields:
        CoreNLPDoc: the annotated document.
    """
    for i, (filename, file_contents) in enumerate(_walk_directory(input_dir)):
        print('Parsing document No. {}: {}'.format(str(i), filename))
        parsed_doc = parser.parse(file_contents, show_tok, show_doc)

        if out_dir:
            out_path = out_dir / (filename + '.json')
            print('Dumping parsed document to {}'.format(out_path))
            with open(out_path, 'w') as file:
                file.write(parsed_doc.pretty_json())
        yield parsed_doc


def _read_lexicon_from_csv(lexicon_path: str) -> dict:
    """Read a lexicon from a CSV file.

    Returns: Dictionary of LexiconElements, each LexiconElement presents a row.
    """
    lexicon = {}
    with open(INFERENCE_LEXICONS / lexicon_path, newline='', encoding="utf-8") as csv_file:
        reader = csv.reader(csv_file, delimiter=',', quotechar='|')
        for row in reader:
            try:
                lexicon[row[0]] = LexiconElement(term=row[0], score=row[1], polarity=None,
                                                 is_acquired=None, position=row[2])
            except IndexError:
                lexicon[row[0]] = LexiconElement(term=row[0], score=row[1], polarity=None,
                                                 is_acquired=None, position=None)
    return lexicon


def load_opinion_lex(file_name) -> dict:
    """Read opinion lexicon from CSV file.

    Returns: Dictionary of LexiconElements, each LexiconElement presents a row.
    """
    lexicon = {}
    with open(file_name, newline='', encoding="utf-8") as csvfile:
        reader = csv.reader(csvfile, delimiter=',', quotechar='|')
        next(reader)
        for row in reader:
            term, score, polarity, is_acquired = row[0], row[1], row[2], row[3]
            score = float(score)
            # ignore terms with low score
            if score >= 0.5 and polarity in (Polarity.POS.value, Polarity.NEG.value):
                lexicon[term] = \
                    LexiconElement(term.lower(),
                                   score if polarity == Polarity.POS.value else -score, polarity,
                                   is_acquired)
    return lexicon


def _load_aspect_lexicon(file_name):
    """Read aspect lexicon from CSV file.

    Returns: Dictionary of LexiconElements, each LexiconElement presents a row.
    """
    lexicon = []
    with open(file_name, 'r', newline='', encoding="utf-8-sig") as csv_file:
        reader = csv.reader(csv_file, delimiter=',', quotechar='|')
        for row in reader:
            lexicon.append(LexiconElement(row))
    return lexicon


def _load_parsed_docs_from_dir(directory: str or PathLike):
    """Read all file in directory.

    Args:
        directory (PathLike): path
    """
    res = {}
    for file_name in listdir(directory):
        if file_name.endswith('.txt') or file_name.endswith('.json'):
            with open(Path(directory) / file_name) as f:
                content = f.read()
                res[file_name] = json.loads(content, object_hook=CoreNLPDoc.decoder)
    return res


def _write_table(table, filename):
    """Write table as csv to file system.

    Args:
        table (list): table to be printed as list
        filename (str): file name
    """
    with open(filename, 'w', newline='') as f:
        writer = csv.writer(f, delimiter=',')
        for row in table:
            writer.writerow(row)


def _write_final_lex(dictionary, file_name):
    """Write generated lex as csv to file.

    Args:
        dictionary (list): list of filtered terms
        file_name (str): file name
    """
    candidate_terms = [["#"] + string_list_headers()]
    term_num = 1
    for candidate_term in dictionary:
        term_row = [int(term_num)] + candidate_term.as_string_list()
        candidate_terms.append(term_row)
        term_num += 1
    _write_table(candidate_terms, file_name)


def _write_generic_sentiment_terms(dictionary, file_name):
    """Write generic sentiment terms as csv to file system."""
    generic_terms = [["Term", "Score", "Polarity"]]
    for generic_term in dictionary.values():
        generic_terms.append(
            [str(generic_term), "1.0", generic_term.polarity.name])
    _write_table(generic_terms, file_name)


def _load_lex_as_list_from_csv(file_name):
    """Load lexicon as list.

    Args:
        file_name (str): input csv file name
    """
    lexicon_table = []
    with open(TRAIN_LEXICONS / file_name, 'r') as f:
        reader = csv.DictReader(f, skipinitialspace=True)
        if reader is None:
            print("file name is None")
            return lexicon_table
        for row in reader:
            term = row['Term'].strip()
            lexicon_table.append(term)

        return lexicon_table


def read_generic_lex_from_file(file_name):
    """Read generic opinion lex for term acquisition.

    Args:
        file_name (string): name of csv file
    """
    with open(file_name) as f:
        reader = csv.DictReader(f)
        dict_list = {}
        for row in reader:
            if row["UsedForAcquisition"] == 'Y':
                key_term = row["Term"] if len(row) > 0 else ""
                terms = []
                if len(row) > 0:
                    terms = key_term.split()
                polarity = Polarity[row["Polarity"]]
                dict_list[key_term] = OpinionTerm(terms, polarity)
    return dict_list


def _read_generic_lex_for_similarity(file_name):
    """Read generic opinion terms for similarity calc from csv file.

    Args:
        file_name (string): name of csv file
    """
    with open(file_name) as f:
        reader = csv.DictReader(f)
        dict_list = {}
        for row in reader:
            if row["UsedForReranking"] == 'Y':
                key_term = row["Term"] if len(row) > 0 else ""
                polarity = row["Polarity"]
                dict_list[key_term] = polarity
    return dict_list



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

from __future__ import absolute_import, division, print_function, unicode_literals

import tensorflow as tf

from nlp_architect.contrib.tensorflow.python.keras.layers.crf import CRF
from nlp_architect.contrib.tensorflow.python.keras.utils.layer_utils import load_model, save_model


class SequenceTagger(object):
    """
    A sequence tagging model for POS and Chunks written in Tensorflow (and Keras) based on the
    paper 'Deep multi-task learning with low level tasks supervised at lower layers'.
    The model has 3 Bi-LSTM layers and outputs POS and Chunk tags.

    Args:
        use_cudnn (bool, optional): use GPU based model (CUDNNA cells)
    """

    def __init__(self, use_cudnn=False):
        self.vocabulary_size = None
        self.num_pos_labels = None
        self.num_chunk_labels = None
        self.char_vocab_size = None
        self.feature_size = None
        self.dropout = None
        self.max_word_len = None
        self.classifier = None
        self.optimizer = None
        self.model = None
        self.use_cudnn = use_cudnn

    def build(self,
              vocabulary_size,
              num_pos_labels,
              num_chunk_labels,
              char_vocab_size=None,
              max_word_len=25,
              feature_size=100,
              dropout=0.5,
              classifier='softmax',
              optimizer=None):
        """
        Build a chunker/POS model

        Args:
            vocabulary_size (int): the size of the input vocabulary
            num_pos_labels (int): the size of of POS labels
            num_chunk_labels (int): the sie of chunk labels
            char_vocab_size (int, optional): character vocabulary size
            max_word_len (int, optional): max characters in a word
            feature_size (int, optional): feature size - determines the embedding/LSTM layer \
                hidden state size
            dropout (float, optional): dropout rate
            classifier (str, optional): classifier layer, 'softmax' for softmax or 'crf' for \
                conditional random fields classifier. default is 'softmax'.
            optimizer (tensorflow.python.training.optimizer.Optimizer, optional): optimizer, if \
                None will use default SGD (paper setup)
        """
        self.vocabulary_size = vocabulary_size
        self.char_vocab_size = char_vocab_size
        self.num_pos_labels = num_pos_labels
        self.num_chunk_labels = num_chunk_labels
        self.max_word_len = max_word_len
        self.feature_size = feature_size
        self.dropout = dropout
        self.classifier = classifier

        word_emb_layer = tf.keras.layers.Embedding(self.vocabulary_size, self.feature_size,
                                                   name='embedding', mask_zero=False)
        word_input = tf.keras.layers.Input(shape=(None,))
        word_embedding = word_emb_layer(word_input)
        input_src = word_input
        features = word_embedding

        # add char input if present
        if self.char_vocab_size is not None:
            char_input = tf.keras.layers.Input(shape=(None, self.max_word_len))
            char_emb_layer = tf.keras.layers.Embedding(self.char_vocab_size, 30,
                                                       name='char_embedding',
                                                       mask_zero=False)
            char_embedding = char_emb_layer(char_input)
            char_embedding = tf.keras.layers.TimeDistributed(
                tf.keras.layers.Conv1D(30, 3, padding='same'))(char_embedding)
            char_embedding = tf.keras.layers.TimeDistributed(tf.keras.layers.GlobalMaxPooling1D())(
                char_embedding)

            input_src = [input_src, char_input]
            features = tf.keras.layers.concatenate([word_embedding, char_embedding])

        rnn_layer_1 = tf.keras.layers.Bidirectional(self._rnn_cell(return_sequences=True))(
            features)
        rnn_layer_2 = tf.keras.layers.Bidirectional(self._rnn_cell(return_sequences=True))(
            rnn_layer_1)
        rnn_layer_3 = tf.keras.layers.Bidirectional(self._rnn_cell(return_sequences=True))(
            rnn_layer_2)

        # outputs
        pos_out = tf.keras.layers.Dense(self.num_pos_labels, activation='softmax',
                                        name='pos_output')(rnn_layer_1)
        losses = {'pos_output': 'categorical_crossentropy'}
        metrics = {'pos_output': 'categorical_accuracy'}

        if 'crf' in self.classifier:
            with tf.device('/cpu:0'):
                chunk_crf = CRF(self.num_chunk_labels, name='chunk_crf')
                rnn_layer_3_dense = tf.keras.layers.Dense(self.num_chunk_labels)(
                    tf.keras.layers.Dropout(self.dropout)(rnn_layer_3))
                chunks_out = chunk_crf(rnn_layer_3_dense)
                losses['chunk_crf'] = chunk_crf.loss
                metrics['chunk_crf'] = chunk_crf.viterbi_accuracy
        else:
            chunks_out = tf.keras.layers.TimeDistributed(
                tf.keras.layers.Dense(self.num_chunk_labels,
                                      activation='softmax'),
                name='chunk_out')(rnn_layer_3)
            losses['chunk_out'] = 'categorical_crossentropy'
            metrics['chunk_out'] = 'categorical_accuracy'

        model = tf.keras.Model(input_src, [pos_out, chunks_out])
        if optimizer is None:
            self.optimizer = tf.keras.optimizers.Adam(0.001, clipnorm=5.)
        else:
            self.optimizer = optimizer
        model.compile(optimizer=self.optimizer,
                      loss=losses,
                      metrics=metrics)
        self.model = model

    def load_embedding_weights(self, weights):
        """
        Load word embedding weights into the model embedding layer

        Args:
            weights (numpy.ndarray): 2D matrix of word weights
        """
        assert self.model is not None, 'Cannot assign weights, apply build() before trying to ' \
                                       'loading embedding weights '
        emb_layer = self.model.get_layer(name='embedding')
        assert emb_layer.output_dim == weights.shape[1], 'embedding vectors shape mismatch'
        emb_layer.set_weights([weights])

    def _rnn_cell(self, **kwargs):
        if self.use_cudnn:
            rnn_cell = tf.keras.layers.CuDNNLSTM(self.feature_size, **kwargs)
        else:
            rnn_cell = tf.keras.layers.LSTM(self.feature_size, **kwargs)
        return rnn_cell

    def fit(self, x, y, batch_size=1, epochs=1, validation_data=None, callbacks=None):
        """
        Fit provided X and Y on built model

        Args:
            x: x samples
            y: y samples
            batch_size (int, optional): batch size per sample
            epochs (int, optional): number of epochs to run before ending training process
            validation_data (optional): x and y samples to validate at the end of the epoch
            callbacks (optional): additional callbacks to run with fitting
        """
        self.model.fit(x=x, y=y, batch_size=batch_size, epochs=epochs,
                       validation_data=validation_data, callbacks=callbacks)

    def predict(self, x, batch_size=1):
        """
        Predict labels given x.

        Args:
            x: samples for inference
            batch_size (int, optional): forward pass batch size

        Returns:
            tuple of numpy arrays of pos and chunk labels
        """
        return self.model.predict(x=x, batch_size=batch_size)

    def save(self, filepath):
        """
        Save the model to disk

        Args:
            filepath (str): file name to save model
        """
        topology = {k: v for k, v in self.__dict__.items()}
        topology.pop('model')
        topology.pop('optimizer')
        topology.pop('use_cudnn')
        save_model(self.model, topology, filepath)

    def load(self, filepath):
        """
        Load model from disk

        Args:
            filepath (str): file name of model
        """
        load_model(filepath, self)


class SequenceChunker(SequenceTagger):
    """
    A sequence Chunker model written in Tensorflow (and Keras) based SequenceTagger model.
    The model uses only the chunking output of the model.
    """

    def predict(self, x, batch_size=1):
        """
        Predict labels given x.

        Args:
            x: samples for inference
            batch_size (int, optional): forward pass batch size

        Returns:
            tuple of numpy arrays of chunk labels
        """
        model = tf.keras.Model(self.model.input, self.model.output[-1])
        return model.predict(x=x, batch_size=batch_size)


class SequencePOSTagger(SequenceTagger):
    """
        A sequence POS tagger model written in Tensorflow (and Keras) based SequenceTagger model.
        The model uses only the chunking output of the model.
        """

    def predict(self, x, batch_size=1):
        """
        Predict labels given x.

        Args:
            x: samples for inference
            batch_size (int, optional): forward pass batch size

        Returns:
            tuple of numpy arrays of POS labels
        """
        model = tf.keras.Model(self.model.input, self.model.output[0])
        return model.predict(x=x, batch_size=batch_size)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import numpy as np


def max_values_squad(data_train):

    """
    Function to compute the maximum length of sentences in
    paragraphs and questions

    Args:
    -----------
    data_train: list containing the entire dataset

    Returns:
    --------
    maximum length of question and paragraph
    """
    max_hypothesis = len(data_train[0][1])
    max_premise = len(data_train[0][0])

    for ele in data_train:
        len1 = len(ele[0])
        len2 = len(ele[1])
        if max_premise < len1:
            max_premise = len1
        if max_hypothesis < len2:
            max_hypothesis = len2

    return max_premise, max_hypothesis


def get_qids(args, q_id_path, data_dev):

    """
    Function to create a list of question_ids in dev set

    Args:
    -----------
    q_id_path: path to question_ids file
    data_dev: development set

    Returns:
    --------
    list of question ids
    """
    qids_list = []
    with open(q_id_path) as q_ids:
        for ele in q_ids:
            qids_list.append(ele.strip().replace(" ", ""))

    final_qidlist = []
    count = 0
    for ele in data_dev:
        para_idx = ele[0]
        if len(para_idx) < args.max_para:
            final_qidlist.append(qids_list[count])
        count += 1

    return final_qidlist


def create_squad_training(paras_file, ques_file, answer_file, data_train_len=None):

    """
    Function to read data from preprocessed files and return
    data in the form of a list

    Args:
    -----------
    paras_file: File name for preprocessed paragraphs
    ques_file: File name for preprocessed questions
    answer_file: File name for preprocessed answer spans
    vocab_file:  File name for preprocessed vocab
    data_train_len= length of train dataset to use

    Returns:
    --------
    appended list for train/dev dataset
    """

    para_list = []
    ques_list = []
    ans_list = []
    with open(paras_file) as f_para:
        for ele in f_para:
            para_list.append(list(map(int, ele.strip().split())))

    with open(ques_file) as f_ques:
        for ele in f_ques:
            ques_list.append(list(map(int, ele.strip().split())))

    with open(answer_file) as f_ans:
        for ele in f_ans:
            ans_list.append(list(map(int, ele.strip().split())))

    data_train = []
    if data_train_len is None:
        data_train_len = len(para_list)

    for idx in range(data_train_len):
        data_train.append([para_list[idx], ques_list[idx], ans_list[idx]])

    return data_train


def get_data_array_squad(params_dict, data_train, set_val='train'):
    """
    Function to pad all sentences and restrict to max length defined by user

    Args:
    ---------
    params_dict: dictionary containing all input parameters
    data_train: list containing the training/dev data_train
    set_val: indicates id its a training set or dev set

    Returns:
    ----------
    Returns a list of tuples with padded sentences and masks
    """

    max_para = params_dict['max_para']
    max_question = params_dict['max_question']
    train_set = []
    count = 0
    for ele in data_train:

        para = ele[0]
        para_idx = ele[0]

        if len(para_idx) < max_para:
            pad_length = max_para - len(para_idx)
            para_idx = para_idx + [0] * pad_length
            para_len = len(para)
            para_mask = np.zeros([1, max_para])
            para_mask[0, 0:len(para)] = 1
            para_mask = para_mask.tolist()[0]

            question_idx = ele[1]
            question = ele[1]

            if len(question) < max_question:
                pad_length = max_question - len(question)
                question_idx = question_idx + [0] * pad_length
                question_len = len(question)
                ques_mask = np.zeros([1, max_question])
                ques_mask[0, 0:question_len] = 1
                ques_mask = ques_mask.tolist()[0]

            train_set.append((para_idx, question_idx, para_len, question_len,
                              ele[2], para_mask, ques_mask))

            if set_val == 'train':
                count += 1
                if count >= params_dict['train_set_size']:
                    break
    return train_set


def create_data_dict(data):

    """
    Function to convert data to dictionary format

    Args:
    -----------
    data: train/dev data as a list

    Returns:
    --------
    a dictionary containing dev/train data
    """

    train = {}
    train['para'] = []
    train['answer'] = []
    train['question'] = []
    train['question_len'] = []
    train['para_len'] = []
    train['para_mask'] = []
    train['question_mask'] = []

    for (para_idx, question_idx, para_len, question_len, answer, para_mask, ques_mask) in data:

        train['para'].append(para_idx)
        train['question'].append(question_idx)
        train['para_len'].append(para_len)
        train['question_len'].append(question_len)
        train['answer'].append(answer)
        train['para_mask'].append(para_mask)
        train['question_mask'].append(ques_mask)

    return train



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import absolute_import
from __future__ import division

import tensorflow as tf
from six.moves import range


def zero_nil_slot(t):
    """
    Overwrites the nil_slot (first row) of the input Tensor with zeros.

    The nil_slot is a dummy slot and should not be trained and influence
    the training algorithm.
    """
    t = tf.convert_to_tensor(t, name="t")
    s = tf.shape(t)[1]
    z = tf.zeros(tf.stack([1, s]))
    return tf.concat(axis=0, values=[z, tf.slice(t, [1, 0], [-1, -1])])


class MemN2N_Dialog(object):
    """End-To-End Memory Network."""
    def __init__(self,
                 batch_size,
                 vocab_size,
                 sentence_size,
                 memory_size,
                 embedding_size,
                 num_cands,
                 max_cand_len,
                 hops=3,
                 max_grad_norm=40.0,
                 nonlin=None,
                 initializer=tf.random_normal_initializer(stddev=0.1),
                 optimizer=tf.train.AdamOptimizer(learning_rate=0.001, epsilon=1e-8),
                 session=tf.Session(),
                 name='MemN2N_Dialog'):
        """Creates an End-To-End Memory Network for Goal Oriented Dialog

        Args:
            cands: Encoded candidate answers

            batch_size: The size of the batch.

            vocab_size: The size of the vocabulary (should include the nil word). The nil word
            one-hot encoding should be 0.

            sentence_size: The max size of a sentence in the data. All sentences should be padded
            to this length. If padding is required it should be done with nil one-hot encoding (0).

            memory_size: The max size of the memory. Since Tensorflow currently does not support
            jagged arrays all memories must be padded to this length. If padding is required, the
            extra memories should be

            empty memories; memories filled with the nil word ([0, 0, 0, ......, 0]).

            embedding_size: The size of the word embedding.

            hops: The number of hops. A hop consists of reading and addressing a memory slot.
            Defaults to `3`.

            max_grad_norm: Maximum L2 norm clipping value. Defaults to `40.0`.

            nonlin: Non-linearity. Defaults to `None`.

            initializer: Weight initializer. Defaults to
            `tf.random_normal_initializer(stddev=0.1)`.

            optimizer: Optimizer algorithm used for SGD. Defaults to
            `tf.train.AdamOptimizer(learning_rate=1e-2)`.

            session: Tensorflow Session the model is run with. Defaults to `tf.Session()`.

            name: Name of the End-To-End Memory Network. Defaults to `MemN2N`.
        """

        self._batch_size = batch_size
        self._vocab_size = vocab_size
        self._sentence_size = sentence_size
        self._memory_size = memory_size
        self._embedding_size = embedding_size
        self._max_cand_len = max_cand_len
        self._num_cands = num_cands
        self._hops = hops
        self._max_grad_norm = max_grad_norm
        self._nonlin = nonlin
        self._init = initializer
        self._name = name

        self._build_inputs()
        self._build_vars()

        self._opt = optimizer

        # cross entropy
        logits = self._inference(self._stories, self._queries)  # (batch_size, vocab_size)
        cross_entropy = tf.nn.softmax_cross_entropy_with_logits(logits=logits,
                                                                labels=tf.cast(self._answers,
                                                                               tf.float32),
                                                                name="cross_entropy")
        # loss op
        cross_entropy_sum = tf.reduce_sum(cross_entropy, name="cross_entropy_sum")

        # gradient pipeline
        grads_and_vars = self._opt.compute_gradients(cross_entropy_sum)
        grads_and_vars = [(tf.clip_by_norm(g, self._max_grad_norm), v) for g, v in grads_and_vars]
        nil_grads_and_vars = []
        for g, v in grads_and_vars:
            if v.name in self._nil_vars:
                nil_grads_and_vars.append((zero_nil_slot(g), v))
            else:
                nil_grads_and_vars.append((g, v))
        train_op = self._opt.apply_gradients(nil_grads_and_vars, name="train_op")

        # predict ops
        predict_op = tf.argmax(logits, 1, name="predict_op")
        predict_proba_op = tf.nn.softmax(logits, name="predict_proba_op")
        predict_log_proba_op = tf.log(predict_proba_op, name="predict_log_proba_op")

        # assign ops
        self.loss_op = cross_entropy_sum
        self.predict_op = predict_op
        self.predict_proba_op = predict_proba_op
        self.predict_log_proba_op = predict_log_proba_op
        self.train_op = train_op

        init_op = tf.global_variables_initializer()
        self._sess = session
        self._sess.run(init_op)
        self.saver = tf.train.Saver(max_to_keep=1)

    def _build_inputs(self):
        self._stories = tf.placeholder(tf.int32, [None, None, self._sentence_size], name="stories")
        self._queries = tf.placeholder(tf.int32, [None, self._sentence_size], name="queries")
        self._answers = tf.placeholder(tf.int32, [None, self._num_cands], name="answers")
        self._cands = tf.placeholder(tf.int32, [None, self._num_cands, self._max_cand_len],
                                     name="candidate_answers")

    def _build_vars(self):
        with tf.variable_scope(self._name):
            nil_word_slot = tf.zeros([1, self._embedding_size])
            A = tf.concat(axis=0, values=[nil_word_slot, self._init([self._vocab_size - 1,
                                                                     self._embedding_size])])
            W = tf.concat(axis=0, values=[nil_word_slot, self._init([self._vocab_size - 1,
                                                                     self._embedding_size])])

            self.LUT_A = tf.Variable(A, name="LUT_A")
            self.LUT_W = tf.Variable(W, name="LUT_W")

            # Dont use projection for layerwise weight sharing
            self.R_proj = tf.Variable(self._init([self._embedding_size, self._embedding_size]),
                                      name="R_proj")

        self._nil_vars = set([self.LUT_A.name, self.LUT_W.name])

    def _inference(self, stories, queries):
        with tf.variable_scope(self._name):
            q_emb = tf.nn.embedding_lookup(self.LUT_A, queries)
            u_0 = tf.reduce_sum(q_emb, 1)
            u = [u_0]

            for _ in range(self._hops):
                m_emb_A = tf.nn.embedding_lookup(self.LUT_A, stories)
                m_A = tf.reduce_sum(m_emb_A, 2)

                # hack to get around no reduce_dot
                u_temp = tf.transpose(tf.expand_dims(u[-1], -1), [0, 2, 1])
                dotted = tf.reduce_sum(m_A * u_temp, 2)

                # Calculate probabilities
                probs = tf.nn.softmax(dotted)

                probs_temp = tf.transpose(tf.expand_dims(probs, -1), [0, 2, 1])

                # Reuse A for the output memory encoding
                c_temp = tf.transpose(m_A, [0, 2, 1])
                o_k = tf.reduce_sum(c_temp * probs_temp, 2)

                # Project hidden state, and add update
                u_k = tf.matmul(u[-1], self.R_proj) + o_k

                # nonlinearity
                if self._nonlin:
                    u_k = self._nonlin(u_k)

                u.append(u_k)

            cands_emb = tf.nn.embedding_lookup(self.LUT_W, self._cands)
            cands_emb_sum = tf.reduce_sum(cands_emb, 2)

            logits = tf.reshape(tf.matmul(tf.expand_dims(u_k, 1),
                                          tf.transpose(cands_emb_sum, [0, 2, 1])),
                                (-1, cands_emb_sum.shape[1]))

            return logits

    def batch_fit(self, stories, queries, answers, cands):
        """Runs the training algorithm over the passed batch

        Args:
            stories: Tensor (None, memory_size, sentence_size)
            queries: Tensor (None, sentence_size)
            answers: Tensor (None, vocab_size)

        Returns:
            loss: floating-point number, the loss computed for the batch
        """
        feed_dict = {self._stories: stories, self._queries: queries,
                     self._answers: answers, self._cands: cands}
        loss, _ = self._sess.run([self.loss_op, self.train_op], feed_dict=feed_dict)

        return loss

    def predict(self, stories, queries, cands):
        """Predicts answers as one-hot encoding.

        Args:
            stories: Tensor (None, memory_size, sentence_size)
            queries: Tensor (None, sentence_size)

        Returns:
            answers: Tensor (None, vocab_size)
        """
        feed_dict = {self._stories: stories, self._queries: queries, self._cands: cands}
        return self._sess.run(self.predict_op, feed_dict=feed_dict)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import csv
from enum import Enum
from os import PathLike

from nlp_architect.models.absa import TRAIN_LEXICONS


class OpinionTerm:
    """Opinion term.

    Attributes:
       terms (list): list of opinion term
        polarity (Polarity): polarity of the sentiment
    """

    def __init__(self, terms, polarity):
        self.terms = terms
        self.polarity = polarity

    def __str__(self):
        return ' '.join(self.terms)


class AspectTerm(object):
    """Aspect term.

    Attributes:
        terms (list): list of terms
        pos (list): list of pos
    """

    def __init__(self, terms, pos):
        """
        Args:
            terms (list): list of terms
            pos (list): list of pos
        """
        self.terms = terms
        self.pos = pos

    def __str__(self):
        return ' '.join(self.terms)

    def __eq__(self, other):
        """
        Override the default equals behavior.
        """
        return self.terms == other.terms and self.pos == other.pos

    @staticmethod
    def from_token(token):
        return AspectTerm([token.text], [token.norm_pos])


class CandidateTerm(object):
    """Candidate opinion term or aspect term.

    Attributes:
        term (list): list of terms
        pos (list): list of pos
        source_term (list): list of related anchor terms
        sentence (str): sentence text this term
        term_polarity (int): term polarity
    """

    def __init__(self, term_a, term_b, sent_text, candidate_term_polarity):
        """
        Args:
            term_a (DepRelationTerm): first term
            term_b (DepRelationTerm): second term
            sent_text (str): sentence text
            candidate_term_polarity (Polarity): term polarity
        """
        self.term = [term_a.text]
        self.pos = [term_a.norm_pos]
        self.source_term = [term_b.text]
        self.sentence = sent_text
        self.term_polarity = candidate_term_polarity

    def __str__(self):
        return ' '.join(self.term)

    def __eq__(self, other):
        if other is None or self.__class__ != other.__class__:
            return False
        if self.term != other.term if self.term is not None else other.term is not None:
            return False
        if self.source_term != other.source_term if self.source_term is not None else \
                other.source_term is not None:
            return False
        return self.sentence == other.sentence if self.sentence is not None else \
            other.sentence is None

    def __ne__(self, other):
        return not self == other


class DepRelation(object):
    """Generic Relation Entry contains the governor, it's dependent and the relation between them.

    Attributes:
        gov (DepRelationTerm): governor
        dep (DepRelationTerm): dependent
        rel (str): relation type between governor and dependent
    """

    def __init__(self, gov=None, dep=None, rel=None):
        self.gov = gov
        self.dep = dep
        rel_split = rel.split(':')
        self.rel = rel_split[0]
        self.subtype = rel_split[1] if len(rel_split) > 1 else None


class RelCategory(Enum):
    SUBJ = {'nsubj', 'nsubjpass', 'csubj', 'csubjpass'}
    MOD = {'amod', 'acl', 'advcl', 'appos', 'neg', 'nmod'}
    OBJ = {'dobj', 'iobj'}


class DepRelationTerm(object):
    """
    Attributes:
        text (str, optional): token text
        lemma (str, optional): token lemma
        pos (str, optional): token pos
        ner (str, optional): token ner
        idx (int, optional): token start index (within the sentence)
    """

    def __init__(self, text=None, lemma=None, pos=None, ner=None, idx=None):
        self.text = text
        self.lemma = lemma
        self.pos = pos
        self.ner = ner
        self.idx = idx
        self.dep_rel_list = []
        self.gov = None

    @property
    def norm_pos(self):
        return normalize_pos(self.text, self.pos)


def string_list_headers():
    return ["CandidateTerm", "Frequency", "Polarity"]


class QualifiedTerm(object):
    """Qualified term - term that is accepted to generated lexicon.

    Attributes:
        term (list): list of terms
        pos (list): list of pos.
        frequency (int): frequency of filtered term in corpus.
        term_polarity (Polarity): term polarity.

    """

    def __init__(self, term, pos, frequency, term_polarity):
        self.term = term
        self.pos = pos
        self.frequency = frequency
        self.term_polarity = term_polarity

    def as_string_list(self):
        return [' '.join(self.term), str(self.frequency),
                self.term_polarity.name]


def load_lex_as_dict_from_csv(file_name: str or PathLike):
    """Read lexicon as dictionary, key = term, value = pos.

    Args:
        file_name: the csv file name
    """
    lexicon_map = {}
    with open(file_name) as f:
        reader = csv.DictReader(f, skipinitialspace=True)
        if reader is None:
            print("file name is None")
            return lexicon_map
        next(reader)
        for row in reader:
            term = row['Term']
            pos = row['POS subtype']

            lexicon_map[term] = pos
    return lexicon_map


class POS(Enum):
    """Part-of-speech labels."""
    ADJ = 1
    ADV = 2
    AUX = 3
    AUX_PAST = 3
    CONJ = 4
    NUM = 5
    DET = 6
    EX = 7
    FW = 8
    IN = 9
    PREP = 10
    LS = 11
    MD = 12
    MD_CERTAIN = 13
    NN = 14
    PROPER_NAME = 15
    POS = 16
    PRON = 17
    PRON_1_S = 18
    PRON_1_P = 19
    PRON_2_S = 20
    PRON_3_S = 21
    PRON_3_P = 22
    PRON_4_S = 23
    POSSPRON_1_S = 24
    POSSPRON_1_P = 25
    POSSPRON_2_S = 26
    POSSPRON_2_P = 27
    POSSPRON_3_S = 28
    POSSPRON_3_P = 29
    POSSPRON_4_S = 30
    POSSPRON_4_P = 31
    RP = 32
    SYM = 33
    TO = 34
    INTERJ = 35
    VB = 36
    VB_PAST = 37
    VB_PRESENT = 38
    VBG = 39
    VBN = 40
    WH_DET = 41
    WH_PROP = 42
    WH_ADV = 43
    PUNCT = 44
    OTHER = 45


PRONOUNS_LIST = load_lex_as_dict_from_csv(TRAIN_LEXICONS / 'PronounsLex.csv')


def normalize_pos(word, in_pos):
    if in_pos is None:
        return POS.OTHER
    if word.lower() in PRONOUNS_LIST and in_pos.startswith("PR"):
        return POS[PRONOUNS_LIST[word.lower()]]
    if in_pos == "CC":
        return POS.CONJ
    if in_pos == "CD":
        return POS.NUM
    if in_pos == "DT":
        return POS.DET
    if in_pos == "EX":
        return POS.EX
    if in_pos == "FW":
        return POS.FW
    if in_pos == "IN":
        return POS.PREP
    if in_pos == "TO":
        return POS.PREP
    if in_pos.startswith("JJ"):
        return POS.ADJ
    if in_pos == "LS":
        return POS.LS
    if in_pos == "MD":
        return POS.MD
    if in_pos.startswith("NN"):
        return POS.NN
    if in_pos == "PDT":
        return POS.DET
    if in_pos == "POS":
        return POS.POS
    if in_pos.startswith("PR"):
        return POS.PRON
    if in_pos.startswith("RB"):
        return POS.ADV
    if in_pos == "RP":
        return POS.RP
    if in_pos == "SYM":
        return POS.SYM
    if in_pos == "UH":
        return POS.INTERJ
    if in_pos.startswith("VB"):
        return POS.VB
    if in_pos == "WDT":
        return POS.WH_DET
    if in_pos.startswith("WP"):
        return POS.WH_PROP
    if in_pos == "WRB":
        return POS.WH_ADV
    return POS.OTHER


class LoadAspectStopLists(object):
    """A Filter holding all generic and general lexicons, can verify if a given term is contained
     in one of the lexicons - hence belongs to one of the generic / general lexicons or is a valid
     term.

    Attributes:
        generic_opinion_lex (dict): generic opinion lexicon
        determiners_lex (dict): determiners lexicon
        general_adjectives_lex (dict): general adjectives lexicon
        generic_quantifiers_lex (dict): generic quantifiers lexicon
        geographical_adjectives_lex (dict): geographical adjectives lexicon
        intensifiers_lex (dict): intensifiers lexicon
        time_adjective_lex (dict): time adjective lexicon
        ordinal_numbers_lex (dict): ordinal numbers lexicon
        prepositions_lex (dict): prepositions lexicon
        pronouns_lex (dict): pronouns lexicon
        colors_lex (dict): colors lexicon
        negation_lex (dict): negation terms lexicon
    """

    def __init__(self, generic_opinion_lex, determiners_lex, general_adjectives_lex,
                 generic_quantifiers_lex, geographical_adjectives_lex, intensifiers_lex,
                 time_adjective_lex, ordinal_numbers_lex, prepositions_lex, pronouns_lex,
                 colors_lex, negation_lex):
        self.generic_opinion_lex = generic_opinion_lex
        self.determiners_lex = determiners_lex
        self.general_adjectives_lex = general_adjectives_lex
        self.generic_quantifiers_lex = generic_quantifiers_lex
        self.geographical_adjectives_lex = geographical_adjectives_lex
        self.intensifiers_lex = intensifiers_lex
        self.time_adjective_lex = time_adjective_lex
        self.ordinal_numbers_lex = ordinal_numbers_lex
        self.prepositions_lex = prepositions_lex
        self.pronouns_lex = pronouns_lex
        self.colors_lex = colors_lex
        self.negation_lex = negation_lex

    def is_in_stop_list(self, term):
        return any(term in lexicon for lexicon in self.__dict__.values())


class LoadOpinionStopLists(object):
    """A Filter holding all generic and general lexicons, can verify if a given term is contained
     in one of the lexicons - hence belongs to one of the generic / general lexicons or is a valid
     term.

    Attributes:
        determiners_lex (dict): determiners lexicon
        general_adjectives_lex (dict): general adjectives lexicon
        generic_quantifiers_lex (dict): generic quantifiers lexicon
        geographical_adjectives_lex (dict): geographical adjectives lexicon
        intensifiers_lex (dict): intensifiers lexicon
        time_adjective_lex (dict): time adjective lexicon
        ordinal_numbers_lex (dict): ordinal numbers lexicon
        prepositions_lex (dict): prepositions lexicon
        colors_lex (dict): colors lexicon
        negation_lex (dict): negation terms lexicon
    """

    def __init__(self, determiners_lex, general_adjectives_lex, generic_quantifiers_lex,
                 geographical_adjectives_lex, intensifiers_lex, time_adjective_lex,
                 ordinal_numbers_lex, prepositions_lex, colors_lex, negation_lex):
        self.determiners_lex = determiners_lex
        self.general_adjectives_lex = general_adjectives_lex
        self.generic_quantifiers_lex = generic_quantifiers_lex
        self.geographical_adjectives_lex = geographical_adjectives_lex
        self.intensifiers_lex = intensifiers_lex
        self.time_adjective_lex = time_adjective_lex
        self.ordinal_numbers_lex = ordinal_numbers_lex
        self.prepositions_lex = prepositions_lex
        self.colors_lex = colors_lex
        self.negation_lex = negation_lex

    def is_in_stop_list(self, term):
        return any(term in lexicon for lexicon in self.__dict__.values())



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import print_function, division

import io
import os
import time

import numpy as np
import scipy
import tensorflow as tf


class Discriminator:

    def __init__(self, input_data, Y, lr_ph):
        self.input_data = input_data
        self.lr_ph = lr_ph
        self.do_ph = tf.placeholder(name="dropout_ph", dtype=tf.float32)
        self.Y = Y
        self.hid_dim = 2048
        # Build Graph
        self._build_network_graph()
        self.disc_cost = None
        self.disc_opt = None
        self.map_opt = None
        self.W = None

    def _build_network_graph(self):
        """
        Builds the basic inference graph for discriminator
        """
        with tf.variable_scope("Discriminator", reuse=tf.AUTO_REUSE):
            w_init = tf.contrib.layers.xavier_initializer()
            noisy_input = tf.nn.dropout(self.input_data, self.do_ph, name="DO1")
            fc1 = tf.layers.dense(noisy_input, self.hid_dim, kernel_initializer=w_init,
                                  activation=tf.nn.leaky_relu, name="Dense1")
            fc2 = tf.layers.dense(fc1, self.hid_dim, kernel_initializer=w_init,
                                  activation=tf.nn.leaky_relu, name="Dense2")
            self.prediction = tf.layers.dense(fc2, 1, kernel_initializer=w_init,
                                              name="Dense_Sig")

    def build_train_graph(self, disc_pred):
        """
        Builds training graph for discriminator
        Arguments:
             disc_pred(object): Discriminator instance
        """
        # Variables in discrimnator scope
        disc_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, "Discriminator")
        # Binary Cross entropy
        disc_entropy = tf.nn.sigmoid_cross_entropy_with_logits(logits=disc_pred, labels=self.Y)
        # Cost
        self.disc_cost = tf.reduce_mean(disc_entropy)
        # Optimizer
        disc_opt = tf.train.GradientDescentOptimizer(self.lr_ph)
        self.disc_opt = disc_opt.minimize(self.disc_cost, var_list=disc_vars)


class Generator:

    def __init__(self, src_ten, tgt_ten, emb_dim, batch_size, smooth_val, lr_ph, beta, vocab_size):
        self.src_ten = src_ten
        self.tgt_ten = tgt_ten
        self.emb_dim = emb_dim
        self.batch_size = batch_size
        self.smooth_val = smooth_val
        self.beta = beta
        self.lr_ph = lr_ph
        self.vocab_size = vocab_size

        # Placeholders
        self.src_ph = tf.placeholder(name="src_ph", shape=[None], dtype=tf.int32)
        self.tgt_ph = tf.placeholder(name="tgt_ph", shape=[None], dtype=tf.int32)

        # Build Graph
        self._build_network_graph()
        ortho_weight = self._build_ortho_graph(self.W)
        self.assign_weight = self._assign_ortho_weight(ortho_weight)
        self.map_opt = None
        self.W = None

    def _build_network_graph(self):
        """
        Builds basic inference graph for generator
        """
        with tf.variable_scope("Generator", reuse=tf.AUTO_REUSE):
            # Look up tables
            self.src_emb = tf.nn.embedding_lookup(self.src_ten, self.src_ph, name="src_lut")
            self.tgt_emb = tf.nn.embedding_lookup(self.tgt_ten, self.tgt_ph, name="tgt_lut")
            # Map them
            self.mapWX = self._mapper(self.src_emb)
            # Concatenate them
            self.X = tf.concat([self.mapWX, self.tgt_emb], 0, name="X")
            # Set target for discriminator
            Y = np.zeros(shape=(2 * self.batch_size, 1), dtype=np.float32)
            # Label smoothing
            Y[:self.batch_size] = 1 - self.smooth_val
            Y[self.batch_size:] = self.smooth_val
            # Convert to tensor
            self.Y = tf.convert_to_tensor(Y, name="Y")

    def build_train_graph(self, disc_pred):
        """
        Builds training graph for generator
        Arguments:
            disc_pred(object): Discriminator instance

        """
        # Variables in Mapper scope
        map_vars = tf.get_collection(tf.GraphKeys.TRAINABLE_VARIABLES, "Generator/Mapper")
        # Binary Cross entropy
        map_entropy = tf.nn.sigmoid_cross_entropy_with_logits(
            logits=disc_pred, labels=(1 - self.Y))
        # Cost
        map_cost = tf.reduce_mean(map_entropy)
        map_opt = tf.train.GradientDescentOptimizer(self.lr_ph)
        self.map_opt = map_opt.minimize(map_cost, var_list=map_vars)

    def _build_ortho_graph(self, W):
        """
        Builds a graph to orthogonalize weight W
        Arguments:
            W (Tensor): Weight in the mapper
        """
        with tf.variable_scope("Ortho", reuse=tf.AUTO_REUSE):
            a = tf.scalar_mul((1 + self.beta), W)  # (1+B)W
            b = tf.matmul(tf.transpose(W), W)  # WWt
            c = tf.matmul(W, b)  # W(W.Wt)
            d = tf.scalar_mul(self.beta, c)  # B(W.Wt)W
            ortho_weight = a - d
            return ortho_weight

    def _assign_ortho_weight(self, ortho_weight):
        """
        Builds a graph to assign weight W after it is orthogonalized
        Arguments:
             ortho_weight(Tensor): Weight after it is orthogonalized
        """
        return tf.assign(self.W, ortho_weight)

    def _mapper(self, src_emb):
        """
        Learns WX mapping to make ||WX-Y|| smaller
        Arguments:
             src_emb(Tensor): Source embeddings after lookup
        """
        with tf.variable_scope("Mapper", reuse=tf.AUTO_REUSE):
            # Initialize as an eye of emb_dim x emb_dim
            self.W = tf.Variable(name="W", initial_value=tf.eye(self.emb_dim, self.emb_dim))
            # Do Matrix Multiply
            WX = tf.matmul(src_emb, self.W)
            # Returns map and weight handles
            return WX


class WordTranslator:
    """
    Main network which does cross-lingual embeddings training
    """

    def __init__(self, hparams, src_vec, tgt_vec, vocab_size):
        # Hyperparameters
        self.batch_size = hparams.batch_size
        self.smooth_val = hparams.smooth_val
        self.beta = hparams.beta
        self.most_freq = hparams.most_freq
        self.emb_dim = hparams.emb_dim
        self.vocab_size = vocab_size
        self.disc_runs = hparams.disc_runs
        self.iters_epoch = hparams.iters_epoch
        self.src_vec = src_vec
        self.tgt_vec = tgt_vec
        self.src_ten = tf.convert_to_tensor(src_vec)
        self.tgt_ten = tf.convert_to_tensor(tgt_vec)
        self.save_dir = hparams.weight_dir
        self.slang = hparams.src_lang
        self.tlang = hparams.tgt_lang

        # Placeholders
        self.lr_ph = tf.placeholder(tf.float32, name="lrPh")
        # Build Graph
        self._build_network_graph()
        self._build_train_graph()

    def _build_network_graph(self):
        """
        Builds inference graph for the GAN
        """
        self.generator = Generator(self.src_ten, self.tgt_ten, self.emb_dim, self.batch_size,
                                   self.smooth_val, self.lr_ph, self.beta, self.vocab_size)
        self.discriminator = Discriminator(self.generator.X, self.generator.Y, self.lr_ph)

    def _build_train_graph(self):
        """
        Builds training graph for the GAN
        """
        self.generator.build_train_graph(self.discriminator.prediction)
        self.discriminator.build_train_graph(self.discriminator.prediction)

    @staticmethod
    def report_metrics(iters, n_words_proc, disc_cost_acc, tic):
        """
        Reports metrics of how training is going
        """
        if iters > 0 and iters % 500 == 0:
            mean_cost = str(sum(disc_cost_acc) / len(disc_cost_acc))
            print(str(int(n_words_proc / (time.time() - tic))) + " Samples/Sec - Iter "
                  + str(iters) + " Discriminator Cost: " + mean_cost)
            # Reset instrumentation
            del disc_cost_acc
            disc_cost_acc = []
            n_words_proc = 0
            tic = time.time()

    def run_generator(self, sess, local_lr):
        """
        Runs generator part of GAN
        Arguments:
            sess(tf.session): Tensorflow Session
            local_lr(float): Learning rate
        Returns:
            Returns number of words processed
        """
        # Generate random ids to look up
        src_ids = np.random.choice(self.vocab_size, self.batch_size, replace=False)
        tgt_ids = np.random.choice(self.vocab_size, self.batch_size, replace=False)
        train_dict = {self.generator.src_ph: src_ids,
                      self.generator.tgt_ph: tgt_ids,
                      self.discriminator.do_ph: 1.0,
                      self.lr_ph: local_lr}
        sess.run(self.generator.map_opt, feed_dict=train_dict)
        # Run orthogonalize
        sess.run(self.generator.assign_weight)
        return 2 * self.batch_size

    def run_discriminator(self, sess, local_lr):
        """
        Runs discriminator part of GAN
        Arguments:
            sess(tf.session): Tensorflow Session
            local_lr(float): Learning rate
        """
        # Generate random ids to look up
        src_ids = np.random.choice(self.most_freq, self.batch_size, replace=False)
        tgt_ids = np.random.choice(self.most_freq, self.batch_size, replace=False)
        train_dict = {self.generator.src_ph: src_ids,
                      self.generator.tgt_ph: tgt_ids,
                      self.discriminator.do_ph: 0.9,
                      self.lr_ph: local_lr}
        return sess.run([self.discriminator.disc_cost, self.discriminator.disc_opt],
                        feed_dict=train_dict)

    def run(self, sess, local_lr):
        """
        Runs whole GAN
        Arguments:
            sess(tf.session): Tensorflow Session
            local_lr(float): Learning rate
        """
        disc_cost_acc = []
        n_words_proc = 0
        tic = time.time()
        for iters in range(0, self.iters_epoch, self.batch_size):
            # 1.Run the discriminator
            for _ in range(self.disc_runs):
                disc_result = self.run_discriminator(sess, local_lr)
                disc_cost_acc.append(disc_result[0])
            # 2.Run the Generator
            n_words_proc += self.run_generator(sess, local_lr)
            # 3.Report the metrics
            self.report_metrics(iters, n_words_proc, disc_cost_acc, tic)

    @staticmethod
    def set_lr(local_lr, drop_lr):
        """
        Drops learning rate based on CSLS criterion
        Arguments:
            local_lr(float): Learning Rate
            drop_lr(bool): Drop learning rate by 2 if True
        """
        new_lr = local_lr * 0.98
        print("Dropping learning rate to " + str(new_lr) + " from " + str(local_lr))
        if drop_lr:
            new_lr = new_lr / 2.0
            print("Dividing learning rate by 2 as validation criterion\
                   decreased. New lr is " + str(new_lr))
        return new_lr

    def save_model(self, save_model, sess):
        """
        Saves W in mapper as numpy array based on CSLS criterion
        Arguments:
            save_model(bool): Save model if True
            sess(tf.session): Tensorflow Session
        """
        if save_model:
            print("Saving model ....")
            model_W = sess.run(self.generator.W)
            path = os.path.join(self.save_dir, "W_best_mapping")
            np.save(path, model_W)

    def apply_procrustes(self, sess, final_pairs):
        """
        Applies procrustes to W matrix for better mapping
        Arguments:
            sess(tf.session): Tensorflow Session
            final_pairs(ndarray): Array of pairs which are mutual neighbors
        """
        print("Applying solution of Procrustes problem to get better mapping...")
        proc_dict = {self.generator.src_ph: final_pairs[:, 0],
                     self.generator.tgt_ph: final_pairs[:, 1]}
        A, B = sess.run([self.generator.src_emb,
                         self.generator.tgt_emb],
                        feed_dict=proc_dict)
        # pylint: disable=no-member
        R = scipy.linalg.orthogonal_procrustes(A, B)
        sess.run(tf.assign(self.generator.W, R[0]))

    def generate_xling_embed(self, sess, src_dict, tgt_dict, tgt_vec):
        """
        Generates cross lingual embeddings
        Arguments:
             sess(tf.session): Tensorflow session
        """
        print("Generating Cross-lingual embeddings...")
        src_emb_x = []
        batch_size = 512
        for i in range(0, self.vocab_size, batch_size):
            sids = [x for x in range(i, min(i + batch_size, self.vocab_size))]
            src_emb_x.append(sess.run(self.generator.mapWX,
                                      feed_dict={self.generator.src_ph: sids}))
        src_emb_x = np.concatenate(src_emb_x)
        print("Writing cross-lingual embeddings to file...")
        src_path = os.path.join(self.save_dir, "vectors-%s.txt" % self.slang)
        tgt_path = os.path.join(self.save_dir, "vectors-%s.txt" % self.tlang)
        with io.open(src_path, "w", encoding="utf-8") as f:
            f.write(u"%i %i\n" % src_emb_x.shape)
            for i in range(len(src_dict)):
                f.write(u"%s %s\n" % (src_dict[i], " ".join('%.5f' % x for x in src_emb_x[i])))

        with io.open(tgt_path, "w", encoding="utf-8") as f:
            f.write(u"%i %i\n" % tgt_vec.shape)
            for i in range(len(tgt_dict)):
                f.write(u"%s %s\n" % (tgt_dict[i], " ".join('%.5f' % x for x in tgt_vec[i])))



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************


def simple_ensembler(np_arrays, weights):
    """
    Simple ensembler takes a list of n by m numpy array predictions and a weight list
    The predictions should be n by m. n is the number of elements and m is the number of classes


    Modified from the default LookupTable implementation to support multiple axis lookups.

    Args:
        vocab_size (int): the vocabulary size
        embed_dim (int): the size of embedding vector
        init (Initializor): initialization function
        update (bool): if the word vectors get updated through training
        pad_idx (int): by knowing the pad value, the update will make sure always
                       have the vector representing pad value to be 0s.
    """
    ensembled_matrix = np_arrays[0] * weights[0]
    for i in range(1, len(np_arrays)):
        ensembled_matrix = ensembled_matrix + np_arrays[i] * weights[i]
    return ensembled_matrix



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import pickle
import tempfile

from tensorflow import keras


def save_model(model: keras.models.Model, topology: dict, filepath: str) -> None:
    """
    Save a model to a file (tf.keras models only)
    The method save the model topology, as given as a
    Args:
        model: model object
        topology (dict): a dictionary of topology elements and their values
        filepath (str): path to save model
    """
    with tempfile.NamedTemporaryFile(suffix='.h5', delete=True) as fd:
        model.save_weights(fd.name)
        model_weights = fd.read()
    data = {'model_weights': model_weights,
            'model_topology': topology}
    with open(filepath, 'wb') as fp:
        pickle.dump(data, fp)


def load_model(filepath, model) -> None:
    """
    Load a model (tf.keras) from disk, create topology from loaded values
    and load weights.
    Args:
        filepath (str): path to model
        model: model object to load
    """
    with open(filepath, 'rb') as fp:
        model_data = pickle.load(fp)
    topology = model_data['model_topology']
    model.build(**topology)
    with tempfile.NamedTemporaryFile(suffix='.h5', delete=True) as fd:
        fd.write(model_data['model_weights'])
        fd.flush()
        model.model.load_weights(fd.name)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from nlp_architect.data.cdc_resources.data_types.wiki.wikipedia_page import WikipediaPage


class WikipediaSearchPageResult(object):
    def __init__(self, search_phrase: str, page_result: WikipediaPage):
        """
        Args:
            search_phrase: the search phrase that yield this page result
            page_result: page result for search phrase
        """
        self.search_phrase = search_phrase
        self.page_result = page_result



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
# Changes Made from original:
#   import paths
# ************************nlp_architect.models.gnmt******************************************************
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
# pylint: skip-file
"""Utility for evaluating various tasks, e.g., translation & summarization."""
import codecs
import os
import re
import subprocess

import tensorflow as tf

from nlp_architect.models.gnmt.scripts import bleu
from nlp_architect.models.gnmt.scripts import rouge

__all__ = ["evaluate"]


def evaluate(ref_file, trans_file, metric, subword_option=None):
    """Pick a metric and evaluate depending on task."""
    # BLEU scores for translation task
    if metric.lower() == "bleu":
        evaluation_score = _bleu(ref_file, trans_file,
                                 subword_option=subword_option)
    # ROUGE scores for summarization tasks
    elif metric.lower() == "rouge":
        evaluation_score = _rouge(ref_file, trans_file,
                                  subword_option=subword_option)
    elif metric.lower() == "accuracy":
        evaluation_score = _accuracy(ref_file, trans_file)
    elif metric.lower() == "word_accuracy":
        evaluation_score = _word_accuracy(ref_file, trans_file)
    else:
        raise ValueError("Unknown metric %s" % metric)

    return evaluation_score


def _clean(sentence, subword_option):
    """Clean and handle BPE or SPM outputs."""
    sentence = sentence.strip()

    # BPE
    if subword_option == "bpe":
        sentence = re.sub("@@ ", "", sentence)

    # SPM
    elif subword_option == "spm":
        sentence = u"".join(sentence.split()).replace(u"\u2581", u" ").lstrip()

    return sentence


# Follow //transconsole/localization/machine_translation/metrics/bleu_calc.py
def _bleu(ref_file, trans_file, subword_option=None):
    """Compute BLEU scores and handling BPE."""
    max_order = 4
    smooth = False

    ref_files = [ref_file]
    reference_text = []
    for reference_filename in ref_files:
        with codecs.getreader("utf-8")(
                tf.gfile.GFile(reference_filename, "rb")) as fh:
            reference_text.append(fh.readlines())

    per_segment_references = []
    for references in zip(*reference_text):
        reference_list = []
        for reference in references:
            reference = _clean(reference, subword_option)
            reference_list.append(reference.split(" "))
        per_segment_references.append(reference_list)

    translations = []
    with codecs.getreader("utf-8")(tf.gfile.GFile(trans_file, "rb")) as fh:
        for line in fh:
            line = _clean(line, subword_option=None)
            translations.append(line.split(" "))

    # bleu_score, precisions, bp, ratio, translation_length, reference_length
    bleu_score, _, _, _, _, _ = bleu.compute_bleu(
        per_segment_references, translations, max_order, smooth)
    return 100 * bleu_score


def _rouge(ref_file, summarization_file, subword_option=None):
    """Compute ROUGE scores and handling BPE."""

    references = []
    with codecs.getreader("utf-8")(tf.gfile.GFile(ref_file, "rb")) as fh:
        for line in fh:
            references.append(_clean(line, subword_option))

    hypotheses = []
    with codecs.getreader("utf-8")(
            tf.gfile.GFile(summarization_file, "rb")) as fh:
        for line in fh:
            hypotheses.append(_clean(line, subword_option=None))

    rouge_score_map = rouge.rouge(hypotheses, references)
    return 100 * rouge_score_map["rouge_l/f_score"]


def _accuracy(label_file, pred_file):
    """Compute accuracy, each line contains a label."""

    with codecs.getreader("utf-8")(tf.gfile.GFile(label_file, "rb")) as label_fh:
        with codecs.getreader("utf-8")(tf.gfile.GFile(pred_file, "rb")) as pred_fh:
            count = 0.0
            match = 0.0
            for label in label_fh:
                label = label.strip()
                pred = pred_fh.readline().strip()
                if label == pred:
                    match += 1
                count += 1
    return 100 * match / count


def _word_accuracy(label_file, pred_file):
    """Compute accuracy on per word basis."""

    with codecs.getreader("utf-8")(tf.gfile.GFile(label_file, "r")) as label_fh:
        with codecs.getreader("utf-8")(tf.gfile.GFile(pred_file, "r")) as pred_fh:
            total_acc, total_count = 0., 0.
            for sentence in label_fh:
                labels = sentence.strip().split(" ")
                preds = pred_fh.readline().strip().split(" ")
                match = 0.0
                for pos in range(min(len(labels), len(preds))):
                    label = labels[pos]
                    pred = preds[pos]
                    if label == pred:
                        match += 1
                total_acc += 100 * match / max(len(labels), len(preds))
                total_count += 1
    return total_acc / total_count


def _moses_bleu(multi_bleu_script, tgt_test, trans_file, subword_option=None):
    """Compute BLEU scores using Moses multi-bleu.perl script."""

    # TODO(thangluong): perform rewrite using python
    # BPE
    if subword_option == "bpe":
        debpe_tgt_test = tgt_test + ".debpe"
        if not os.path.exists(debpe_tgt_test):
            # TODO(thangluong): not use shell=True, can be a security hazard
            cp_cmd = "cp %s %s" % (tgt_test, debpe_tgt_test)
            subprocess.call(cp_cmd.split(), shell=False)
            sed_cmd = "sed s/@@ //g %s" % (debpe_tgt_test)
            subprocess.call(sed_cmd.split(), shell=False)
        tgt_test = debpe_tgt_test
    elif subword_option == "spm":
        despm_tgt_test = tgt_test + ".despm"
        if not os.path.exists(despm_tgt_test):
            subprocess.call("cp %s %s" % (tgt_test, despm_tgt_test))
            subprocess.call("sed s/ //g %s" % (despm_tgt_test))
            subprocess.call(u"sed s/^\u2581/g %s" % (despm_tgt_test))
            subprocess.call(u"sed s/\u2581/ /g %s" % (despm_tgt_test))
        tgt_test = despm_tgt_test
    cmd = "%s %s < %s" % (multi_bleu_script, tgt_test, trans_file)

    # subprocess
    # TODO(thangluong): not use shell=True, can be a security hazard
    bleu_output = subprocess.check_output(cmd.split(), shell=False)

    # extract BLEU score
    m = re.search("BLEU = (.+?),", bleu_output)
    bleu_score = float(m.group(1))

    return bleu_score



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************

import os
import sys

# Things that were changed from the original:
# - Added legal header
# - Reformatted code and variable names to conform with PEP8
# - Added pointer to 'weights.clas' file
# - Added run_conllu_eval()
# - Removed tests and command-line usage option
# - Removed unnecessary imports
# - Add pylint check disable flags

# !/usr/bin/env python
# CoNLL 2017 UD Parsing evaluation script.
#
# Compatible with Python 2.7 and 3.2+, can be used either as a module
# or a standalone executable.
#
# Copyright 2017 Institute of Formal and Applied Linguistics (UFAL),
# Faculty of Mathematics and Physics, Charles University, Czech Republic.
#
# Changelog:
# - [02 Jan 2017] Version 0.9: Initial release
# - [25 Jan 2017] Version 0.9.1: Fix bug in LCS alignment computation
# - [10 Mar 2017] Version 1.0: Add documentation and test
#                              Compare HEADs correctly using aligned words
#                              Allow evaluation with errorneous spaces in forms
#                              Compare forms in LCS case insensitively
#                              Detect cycles and multiple root nodes
#                              Compute AlignedAccuracy
# API usage
# ---------
# - load_conllu(file)
#   - loads CoNLL-U file from given file object to an internal representation
#   - the file object should return str on both Python 2 and Python 3
#   - raises UDError exception if the given file cannot be loaded
# - evaluate(gold_ud, system_ud)
#   - evaluate the given gold and system CoNLL-U files (loaded with
#     load_conllu)
#   - raises UDError if the concatenated tokens of gold and system file do not
#     match
#   - returns a dictionary with the metrics described above, each metrics
#     having three fields: precision, recall and f1
#
# Description of token matching
# -----------------------------
# In order to match tokens of gold file and system file, we consider the text
# resulting from concatenation of gold tokens and text resulting from
# concatenation of system tokens. These texts should match -- if they do not,
# the evaluation fails.
#
# If the texts do match, every token is represented as a range in this original
# text, and tokens are equal only if their range is the same.
#
# Description of word matching
# ----------------------------
# When matching words of gold file and system file, we first match the tokens.
# The words which are also tokens are matched as tokens, but words in
# multi-word tokens have to be handled differently.
#
# To handle multi-word tokens, we start by finding "multi-word spans".
# Multi-word span is a span in the original text such that
# - it contains at least one multi-word token
# - all multi-word tokens in the span (considering both gold and system ones)
#   are completely inside the span (i.e., they do not "stick out")
# - the multi-word span is as small as possible
#
# For every multi-word span, we align the gold and system words completely
# inside this span using LCS on their FORMs. The words not intersecting
# (even partially) any multi-word span are then aligned as tokens.
# pylint: disable=too-many-statements

# CoNLL-U column names
ID, FORM, LEMMA, UPOS, XPOS, FEATS, HEAD, DEPREL, DEPS, MISC = list(range(10))

WEIGHTS = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'weights.clas')


# UD Error is used when raising exceptions in this module
class UDError(Exception):
    pass


# Load given CoNLL-U file into internal representation
def load_conllu(file):
    # pylint: disable=too-many-locals
    # pylint: disable=too-many-branches
    # pylint: disable=too-many-statements

    # Internal representation classes
    class UDRepresentation:
        # pylint: disable=too-few-public-methods
        def __init__(self):
            # Characters of all the tokens in the whole file.
            # Whitespace between tokens is not included.
            self.characters = []
            # List of UDSpan instances with start&end indices into `characters`
            self.tokens = []
            # List of UDWord instances.
            self.words = []
            # List of UDSpan instances with start&end indices into `characters`
            self.sentences = []

    class UDSpan:
        # pylint: disable=too-few-public-methods
        def __init__(self, start, end):
            self.start = start
            # Note that self.end marks the first position **after the end** of
            # span, so we can use characters[start:end] or range(start, end).
            self.end = end

    class UDWord:
        # pylint: disable=too-few-public-methods
        def __init__(self, span, columns, is_multiword):
            # Span of this word (or MWT, see below) within
            # ud_representation.characters.
            self.span = span
            # 10 columns of the CoNLL-U file: ID, FORM, LEMMA,...
            self.columns = columns
            # is_multiword==True means that this word is part of a multi-word
            # token.
            # In that case, self.span marks the span of the whole multi-word
            # token.
            self.is_multiword = is_multiword
            # Reference to the UDWord instance representing the HEAD (or None
            # if root).
            self.parent = None
            # Let's ignore language-specific deprel subtypes.
            self.columns[DEPREL] = columns[DEPREL].split(':')[0]

    ud = UDRepresentation()

    # Load the CoNLL-U file
    index, sentence_start = 0, None
    while True:
        line = file.readline()
        if not line:
            break
        line = line.rstrip("\r\n")

        # Handle sentence start boundaries
        if sentence_start is None:
            # Skip comments
            if line.startswith("#"):
                continue
            # Start a new sentence
            ud.sentences.append(UDSpan(index, 0))
            sentence_start = len(ud.words)
        if not line:
            # Add parent UDWord links and check there are no cycles
            def process_word(word):
                if word.parent == "remapping":
                    raise UDError("There is a cycle in a sentence")
                if word.parent is None:
                    head = int(word.columns[HEAD])
                    if head > len(ud.words) - sentence_start:
                        raise UDError(
                            "HEAD '{}' points outside of the sentence".format(
                                word.columns[HEAD]))
                    if head:
                        parent = ud.words[sentence_start + head - 1]
                        word.parent = "remapping"
                        process_word(parent)
                        word.parent = parent

            for word in ud.words[sentence_start:]:
                process_word(word)

            # Check there is a single root node
            if len([word for word in ud.words[sentence_start:] if
                    word.parent is None]) != 1:
                raise UDError("There are multiple roots in a sentence")

            # End the sentence
            ud.sentences[-1].end = index
            sentence_start = None
            continue

        # Read next token/word
        columns = line.split("\t")
        if len(columns) != 10:
            raise UDError(
                "The CoNLL-U line does not contain 10 tab-separated columns: "
                "'{}'".format(line))

        # Skip empty nodes
        if "." in columns[ID]:
            continue

        # Delete spaces from FORM  so gold.characters == system.characters
        # even if one of them tokenizes the space.
        columns[FORM] = columns[FORM].replace(" ", "")
        if not columns[FORM]:
            raise UDError("There is an empty FORM in the CoNLL-U file")

        # Save token
        ud.characters.extend(columns[FORM])
        ud.tokens.append(UDSpan(index, index + len(columns[FORM])))
        index += len(columns[FORM])

        # Handle multi-word tokens to save word(s)
        if "-" in columns[ID]:
            try:
                start, end = list(map(int, columns[ID].split("-")))
            except Exception:
                raise UDError("Cannot parse multi-word token ID '{}'".format(
                    columns[ID]))

            for _ in range(start, end + 1):
                word_line = file.readline().rstrip("\r\n")
                word_columns = word_line.split("\t")
                if len(word_columns) != 10:
                    raise UDError(
                        "The CoNLL-U line does not contain 10 tab-separated "
                        "columns: '{}'".format(word_line))
                ud.words.append(
                    UDWord(ud.tokens[-1], word_columns, is_multiword=True))
        # Basic tokens/words
        else:
            try:
                word_id = int(columns[ID])
            except Exception:
                raise UDError("Cannot parse word ID '{}'".format(columns[ID]))
            if word_id != len(ud.words) - sentence_start + 1:
                raise UDError("Incorrect word ID '{}' for word '{}', expected"
                              " '{}'".format(columns[ID], columns[FORM],
                                             len(ud.words)
                                             - sentence_start + 1))

            try:
                head_id = int(columns[HEAD])
            except Exception:
                raise UDError("Cannot parse HEAD '{}'".format(columns[HEAD]))
            if head_id < 0:
                raise UDError("HEAD cannot be negative")

            ud.words.append(UDWord(ud.tokens[-1], columns, is_multiword=False))

    if sentence_start is not None:
        raise UDError("The CoNLL-U file does not end with empty line")

    return ud


# Evaluate the gold and system treebanks (loaded using load_conllu).
def evaluate(gold_ud, system_ud, deprel_weights=None):
    # pylint: disable=too-many-locals
    class Score:
        # pylint: disable=too-few-public-methods
        def __init__(self, gold_total, system_total, correct,
                     aligned_total=None):
            self.precision = correct / system_total if system_total else 0.0
            self.recall = correct / gold_total if gold_total else 0.0
            self.f1 = 2 * correct / (system_total + gold_total) \
                if system_total + gold_total else 0.0
            self.aligned_accuracy = \
                correct / aligned_total if aligned_total else aligned_total

    class AlignmentWord:
        # pylint: disable=too-few-public-methods
        def __init__(self, gold_word, system_word):
            self.gold_word = gold_word
            self.system_word = system_word
            self.gold_parent = None
            self.system_parent_gold_aligned = None

    class Alignment:
        def __init__(self, gold_words, system_words):
            self.gold_words = gold_words
            self.system_words = system_words
            self.matched_words = []
            self.matched_words_map = {}

        def append_aligned_words(self, gold_word, system_word):
            self.matched_words.append(AlignmentWord(gold_word, system_word))
            self.matched_words_map[system_word] = gold_word

        def fill_parents(self):
            # We represent root parents in both gold and system data by '0'.
            # For gold data, we represent non-root parent by corresponding gold
            # word.
            # For system data, we represent non-root parent by either gold word
            # aligned
            # to parent system nodes, or by None if no gold words is aligned to
            # the parent.
            for words in self.matched_words:
                words.gold_parent = words.gold_word.parent if \
                    words.gold_word.parent is not None else 0
                words.system_parent_gold_aligned = self.matched_words_map.get(
                    words.system_word.parent, None) \
                    if words.system_word.parent is not None else 0

    def lower(text):
        if sys.version_info < (3, 0) and isinstance(text, str):
            return text.decode("utf-8").lower()
        return text.lower()

    def spans_score(gold_spans, system_spans):
        correct, gi, si = 0, 0, 0
        while gi < len(gold_spans) and si < len(system_spans):
            if system_spans[si].start < gold_spans[gi].start:
                si += 1
            elif gold_spans[gi].start < system_spans[si].start:
                gi += 1
            else:
                correct += gold_spans[gi].end == system_spans[si].end
                si += 1
                gi += 1

        return Score(len(gold_spans), len(system_spans), correct)

    def alignment_score(alignment, key_fn, weight_fn=lambda w: 1):
        gold, system, aligned, correct = 0, 0, 0, 0

        for word in alignment.gold_words:
            gold += weight_fn(word)

        for word in alignment.system_words:
            system += weight_fn(word)

        for words in alignment.matched_words:
            aligned += weight_fn(words.gold_word)

        if key_fn is None:
            # Return score for whole aligned words
            return Score(gold, system, aligned)

        for words in alignment.matched_words:
            if key_fn(words.gold_word, words.gold_parent) == key_fn(
                    words.system_word,
                    words.system_parent_gold_aligned):
                correct += weight_fn(words.gold_word)

        return Score(gold, system, correct, aligned)

    def beyond_end(words, i, multiword_span_end):
        if i >= len(words):
            return True
        if words[i].is_multiword:
            return words[i].span.start >= multiword_span_end
        return words[i].span.end > multiword_span_end

    def extend_end(word, multiword_span_end):
        if word.is_multiword and word.span.end > multiword_span_end:
            return word.span.end
        return multiword_span_end

    def find_multiword_span(gold_words, system_words, gi, si):
        # We know gold_words[gi].is_multiword or system_words[si].is_multiword.
        # Find the start of the multiword span (gs, ss), so the multiword span
        # is minimal.
        # Initialize multiword_span_end characters index.
        if gold_words[gi].is_multiword:
            multiword_span_end = gold_words[gi].span.end
            if not system_words[si].is_multiword and system_words[si].span.start < \
                    gold_words[gi].span.start:
                si += 1
        else:  # if system_words[si].is_multiword
            multiword_span_end = system_words[si].span.end
            if not gold_words[gi].is_multiword and gold_words[gi].span.start < \
                    system_words[si].span.start:
                gi += 1
        gs, ss = gi, si

        # Find the end of the multiword span
        # (so both gi and si are pointing to the word following the multiword
        # span end).
        while not beyond_end(gold_words, gi, multiword_span_end) or \
                not beyond_end(system_words, si, multiword_span_end):
            gold_start = gold_words[gi].span.start
            sys_start = system_words[si].span.start
            if gi < len(gold_words) and (si >= len(system_words) or gold_start <= sys_start):
                multiword_span_end = extend_end(gold_words[gi], multiword_span_end)
                gi += 1
            else:
                multiword_span_end = extend_end(system_words[si], multiword_span_end)
                si += 1
        return gs, ss, gi, si

    def compute_lcs(gold_words, system_words, gi, si, gs, ss):
        # pylint: disable=too-many-arguments
        lcs = [[0] * (si - ss) for _ in range(gi - gs)]
        for g in reversed(list(range(gi - gs))):
            for s in reversed(list(range(si - ss))):
                if lower(gold_words[gs + g].columns[FORM]) == lower(
                        system_words[ss + s].columns[FORM]):
                    lcs[g][s] = 1 + (lcs[g + 1][s + 1] if
                                     g + 1 < gi - gs and s + 1 < si - ss
                                     else 0)
                lcs[g][s] = max(lcs[g][s],
                                lcs[g + 1][s] if g + 1 < gi - gs else 0)
                lcs[g][s] = max(lcs[g][s],
                                lcs[g][s + 1] if s + 1 < si - ss else 0)
        return lcs

    def align_words(gold_words, system_words):
        alignment = Alignment(gold_words, system_words)

        gi, si = 0, 0
        while gi < len(gold_words) and si < len(system_words):
            if gold_words[gi].is_multiword or system_words[si].is_multiword:
                # A: Multi-word tokens => align via LCS within the whole
                # "multiword span".
                gs, ss, gi, si = find_multiword_span(gold_words, system_words,
                                                     gi, si)

                if si > ss and gi > gs:
                    lcs = compute_lcs(gold_words, system_words, gi, si, gs, ss)

                    # Store aligned words
                    s, g = 0, 0
                    while g < gi - gs and s < si - ss:
                        if lower(gold_words[gs + g].columns[FORM]) == lower(
                                system_words[ss + s].columns[FORM]):
                            alignment.append_aligned_words(gold_words[gs + g],
                                                           system_words[
                                                               ss + s])
                            g += 1
                            s += 1
                        elif lcs[g][s] == (
                                lcs[g + 1][s] if g + 1 < gi - gs else 0):
                            g += 1
                        else:
                            s += 1
            else:
                # B: No multi-word token => align according to spans.
                if (gold_words[gi].span.start, gold_words[gi].span.end) == (
                        system_words[si].span.start,
                        system_words[si].span.end):
                    alignment.append_aligned_words(gold_words[gi],
                                                   system_words[si])
                    gi += 1
                    si += 1
                elif gold_words[gi].span.start <= system_words[si].span.start:
                    gi += 1
                else:
                    si += 1

        alignment.fill_parents()

        return alignment

    # Check that underlying character sequences do match
    if gold_ud.characters != system_ud.characters:
        index = 0
        while gold_ud.characters[index] == system_ud.characters[index]:
            index += 1

        raise UDError(
            "The concatenation of tokens in gold file and in system file "
            "differ!\n"
            + "First 20 differing characters in gold file: '{}' and system file:"
            " '{}'".format(
                "".join(gold_ud.characters[index:index + 20]),
                "".join(system_ud.characters[index:index + 20])
            )
        )

    # Align words
    alignment = align_words(gold_ud.words, system_ud.words)

    # Compute the F1-scores
    result = {
        "Tokens": spans_score(gold_ud.tokens, system_ud.tokens),
        "Sentences": spans_score(gold_ud.sentences, system_ud.sentences),
        "Words": alignment_score(alignment, None),
        "UPOS": alignment_score(alignment, lambda w, parent: w.columns[UPOS]),
        "XPOS": alignment_score(alignment, lambda w, parent: w.columns[XPOS]),
        "Feats": alignment_score(alignment,
                                 lambda w, parent: w.columns[FEATS]),
        "AllTags": alignment_score(alignment, lambda w, parent: (
            w.columns[UPOS], w.columns[XPOS], w.columns[FEATS])),
        "Lemmas": alignment_score(alignment,
                                  lambda w, parent: w.columns[LEMMA]),
        "UAS": alignment_score(alignment, lambda w, parent: parent),
        "LAS": alignment_score(alignment,
                               lambda w, parent: (parent, w.columns[DEPREL])),
    }

    # Add WeightedLAS if weights are given
    if deprel_weights is not None:
        def weighted_las(word):
            return deprel_weights.get(word.columns[DEPREL], 1.0)

        result["WeightedLAS"] = alignment_score(alignment, lambda w, parent: (
            parent, w.columns[DEPREL]), weighted_las)

    return result


def load_deprel_weights(weights_file):
    if weights_file is None:
        return None

    deprel_weights = {}
    with open(weights_file) as f:
        for line in f:
            # Ignore comments and empty lines
            if line.startswith("#") or not line.strip():
                continue

            columns = line.rstrip("\r\n").split()
            if len(columns) != 2:
                raise ValueError(
                    "Expected two columns in the UD Relations weights file on line"
                    " '{}'".format(
                        line))

            deprel_weights[columns[0]] = float(columns[1])

    return deprel_weights


def load_conllu_file(path):
    with open(path, mode="r", **({"encoding": "utf-8"} if sys.version_info >= (3, 0) else {})) \
            as _file:
        return load_conllu(_file)


def evaluate_wrapper(gold_file: str, system_file: str, weights_file: str):
    # Load CoNLL-U files
    gold_ud = load_conllu_file(gold_file)
    system_ud = load_conllu_file(system_file)

    # Load weights if requested
    deprel_weights = load_deprel_weights(weights_file)

    return evaluate(gold_ud, system_ud, deprel_weights)


def run_conllu_eval(gold_file, test_file, weights_file=WEIGHTS, verbose=True):
    # Use verbose if weights are supplied
    if weights_file is not None and not verbose:
        verbose = True

    # Evaluate
    evaluation = evaluate_wrapper(gold_file, test_file, weights_file)

    # Write the evaluation to file
    with open(test_file[:test_file.rindex('.')] + '_eval.txt', 'w') as out_file:
        if not verbose:
            out_file.write("LAS F1 Score: {:.2f}".format(100 * evaluation["LAS"].f1) + '\n')
        else:
            metrics = ["Tokens", "Sentences", "Words", "UPOS", "XPOS", "Feats",
                       "AllTags", "Lemmas", "UAS", "LAS"]
            if weights_file is not None:
                metrics.append("WeightedLAS")

            out_file.write("Metrics    | Precision |    Recall |  F1 Score | AligndAcc" + '\n')
            out_file.write("-----------+-----------+-----------+-----------+-----------" + '\n')
            for metric in metrics:
                out_file.write("{:11}|{:10.2f} |{:10.2f} |{:10.2f} |{}".format(
                    metric,
                    100 * evaluation[metric].precision,
                    100 * evaluation[metric].recall,
                    100 * evaluation[metric].f1,
                    "{:10.2f}".format(100 * evaluation[metric].aligned_accuracy)
                    if evaluation[metric].aligned_accuracy is not None else ""
                ) + '\n')



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from __future__ import absolute_import

import numpy as np
import tensorflow as tf
import tensorflow_hub as hub

from nlp_architect.utils.text import Vocabulary


def load_word_embeddings(file_path, vocab=None):
    """
    Loads a word embedding model text file into a word(str) to numpy vector dictionary

    Args:
        file_path (str): path to model file
        vocab (list of str): optional - vocabulary

    Returns:
        list: a dictionary of numpy.ndarray vectors
        int: detected word embedding vector size
    """
    with open(file_path, encoding='utf-8') as fp:
        word_vectors = {}
        size = None
        for line in fp:
            line_fields = line.split()
            if len(line_fields) < 5:
                continue
            else:
                if line[0] == ' ':
                    word_vectors[' '] = np.asarray(line_fields, dtype='float32')
                elif vocab is None or line_fields[0] in vocab:
                    word_vectors[line_fields[0]] = np.asarray(line_fields[1:], dtype='float32')
                    if size is None:
                        size = len(line_fields[1:])
    return word_vectors, size


def fill_embedding_mat(src_mat, src_lex, emb_lex, emb_size):
    """
    Creates a new matrix from given matrix of int words using the embedding
    model provided.

    Args:
        src_mat (numpy.ndarray): source matrix
        src_lex (dict): source matrix lexicon
        emb_lex (dict): embedding lexicon
        emb_size (int): embedding vector size
    """
    emb_mat = np.zeros((src_mat.shape[0], src_mat.shape[1], emb_size))
    for i, sen in enumerate(src_mat):
        for j, w in enumerate(sen):
            if w > 0:
                w_emb = emb_lex.get(str(src_lex.get(w)).lower())
                if w_emb is not None:
                    emb_mat[i][j] = w_emb
    return emb_mat


def get_embedding_matrix(embeddings: dict, vocab: Vocabulary) -> np.ndarray:
    """
    Generate a matrix of word embeddings given a vocabulary

    Args:
        embeddings (dict): a dictionary of embedding vectors
        vocab (Vocabulary): a Vocabulary

    Returns:
        a 2D numpy matrix of lexicon embeddings
    """
    emb_size = len(next(iter(embeddings.values())))
    mat = np.zeros((vocab.max, emb_size))
    for word, wid in vocab.vocab.items():
        vec = embeddings.get(word.lower(), None)
        if vec is not None:
            mat[wid] = vec
    return mat


# pylint: disable=not-context-manager
class ELMoEmbedderTFHUB(object):
    def __init__(self):
        self.g = tf.Graph()

        with self.g.as_default():
            text_input = tf.placeholder(dtype=tf.string)
            text_input_size = tf.placeholder(dtype=tf.int32)
            print('Loading Tensorflow hub ELMo model, '
                  'might take a while on first load (downloading from web)')
            self.elmo = hub.Module("https://tfhub.dev/google/elmo/2", trainable=False)
            self.inputs = {
                'tokens': text_input,
                'sequence_len': text_input_size
            }
            self.embedding = self.elmo(inputs=self.inputs,
                                       signature='tokens',
                                       as_dict=True)['elmo']

            sess = tf.Session(graph=self.g)
            sess.run(tf.global_variables_initializer())
            sess.run(tf.tables_initializer())
            self.s = sess

    def get_vector(self, tokens):
        vec = self.s.run(self.embedding,
                         feed_dict={self.inputs['tokens']: [tokens],
                                    self.inputs['sequence_len']: [len(tokens)]})
        return np.squeeze(vec, axis=0)



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from os import path
from pathlib import Path

from nlp_architect import LIBRARY_OUT

ABSA_ROOT = Path(path.realpath(__file__)).parent

TRAIN_LEXICONS = ABSA_ROOT / 'train' / 'lexicons'

TRAIN_CONF = ABSA_ROOT / 'train' / 'config.ini'

TRAIN_OUT = LIBRARY_OUT / 'absa' / 'train'

INFERENCE_LEXICONS = ABSA_ROOT / 'inference' / 'lexicons'

INFERENCE_OUT = LIBRARY_OUT / 'absa' / 'inference'



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
import logging
from typing import Tuple

from nlp_architect.common.cdc.cluster import Cluster
from nlp_architect.data.cdc_resources.relations.relation_extraction import RelationExtraction
from nlp_architect.data.cdc_resources.relations.relation_types_enums import RelationType

logger = logging.getLogger(__name__)


class SieveClusterMerger(object):
    def __init__(self, excepted_relation: Tuple[RelationType, float],
                 relation_extractor: RelationExtraction):
        """
        Args:
            excepted_relation: tuple with relation to run in sieve,
            threshold to merge clusters
            relation_extractor:
        """
        self.excepted_relation = excepted_relation[0]
        self.threshold = excepted_relation[1]
        self.relation_extractor = relation_extractor

        logger.info('init Sieve, for relation-%s with threshold=%.1f',
                    self.excepted_relation.name, self.threshold)

    def run_sieve(self, cluster_i: Cluster, cluster_j: Cluster) -> bool:
        """
        Args:
            cluster_i:
            cluster_j:

        Returns:
            bool -> indicating whether to merge clusters (True) or not (False)
        """
        matches = 0
        for mention_i in cluster_i.mentions:
            for mention_j in cluster_j.mentions:
                match_result = self.relation_extractor.extract_sub_relations(
                    mention_i, mention_j, self.excepted_relation)
                if match_result == self.excepted_relation:
                    matches += 1

        possible_pairs_len = float(len(cluster_i.mentions) * len(cluster_j.mentions))
        matches_rate = matches / possible_pairs_len

        result = False
        if matches_rate >= self.threshold:
            result = True

        return result



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************


class WikipediaPages(object):
    def __init__(self):
        """
        Object represent a set of Wikipedia Pages
        """
        self.pages = set()
        self.is_empty_norm_phrase = True

    def get_pages(self):
        return self.pages

    def add_page(self, page):
        self.pages.add(page)
        if page.orig_phrase_norm is not None and page.orig_phrase_norm != '':
            self.is_empty_norm_phrase = False

    def get_and_set_all_disambiguation(self):
        all_disambiguations = []
        for page in self.pages:
            if page.relations.disambiguation_links_norm is not None:
                all_disambiguations.extend(page.relations.disambiguation_links_norm)
            if page.relations.disambiguation_links is not None:
                all_disambiguations.extend(page.relations.disambiguation_links)
        return set(all_disambiguations)

    def get_and_set_all_categories(self):
        all_categories = []
        for page in self.pages:
            if page.relations.categories_norm is not None:
                all_categories.extend(page.relations.categories_norm)
            if page.relations.categories is not None:
                all_categories.extend(page.relations.categories)
        return set(all_categories)

    def get_and_set_all_aliases(self):
        all_aliases = []
        for page in self.pages:
            if page.relations.aliases_norm is not None:
                all_aliases.extend(page.relations.aliases_norm)
            if page.relations.aliases is not None:
                all_aliases.extend(page.relations.aliases)
        return set(all_aliases)

    def get_and_set_parenthesis(self):
        all_parenthesis = []
        for page in self.pages:
            if page.relations.title_parenthesis_norm is not None:
                all_parenthesis.extend(page.relations.title_parenthesis_norm)
            if page.relations.title_parenthesis is not None:
                all_parenthesis.extend(page.relations.title_parenthesis)
        return set(all_parenthesis)

    def get_and_set_be_comp(self):
        all_be_comp = []
        for page in self.pages:
            if page.relations.be_comp_norm is not None:
                all_be_comp.extend(page.relations.be_comp_norm)
            if page.relations.be_comp is not None:
                all_be_comp.extend(page.relations.be_comp)
        return set(all_be_comp)

    def get_and_set_titles(self):
        all_titles = []
        for page in self.pages:
            if page.orig_phrase != '':
                all_titles.append(page.orig_phrase)
                all_titles.append(page.orig_phrase_norm)
            if page.wiki_title != '':
                all_titles.append(page.wiki_title)
                all_titles.append(page.wiki_title_norm)
        return set(all_titles)

    def toJson(self):
        result_dict = {}
        page_list = []
        for page in self.pages:
            page_list.append(page.toJson())

        result_dict['pages'] = page_list
        return result_dict

    def __str__(self) -> str:
        result_str = ''
        for page in self.pages:
            result_str += str(page) + ', '

        return result_str.strip()



# ******************************************************************************
# Copyright 2017-2018 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ******************************************************************************
from nlp_architect.models.absa.inference.data_types import Polarity
from nlp_architect.models.absa.train.data_types import OpinionTerm, QualifiedTerm


def set_opinion_term_polarity(terms_list):
    """Set opinion term polarity.

    Args:
        terms_list (list): list of opinion terms
    """
    first = terms_list[0]
    new_term = first.term

    positive_pol = 0
    negative_pol = 0
    curr_polarity = None
    for term in terms_list:
        try:
            curr_polarity = term.term_polarity
        except Exception as e:
            print("extracted_term missing term_polarity: " + str(e))
        if curr_polarity is not None:
            if curr_polarity == Polarity.POS:
                positive_pol = positive_pol + 1
            if curr_polarity == Polarity.NEG:
                negative_pol = negative_pol + 1
    new_term_polarity = Polarity.UNK
    if positive_pol >= negative_pol and positive_pol > 0:
        new_term_polarity = Polarity.POS
    elif negative_pol >= positive_pol and negative_pol > 0:
        new_term_polarity = Polarity.NEG

    return OpinionTerm(new_term, new_term_polarity)


def generate_final_aspect_candidates_list(aspect_candidate_list_raw,
                                          final_aspect_candidates_list,
                                          frequency_threshold):
    """Generate final aspect candidates list from map.

    Args:
        aspect_candidate_list_raw (dict): key = term, value =
        lists of candidate terms.
        final_aspect_candidates_list (list): list of final aspect candidates
        frequency_threshold (int): minimum freq. for qualifying term
    """
    term_polarity = Polarity.UNK
    for extracted_term_list in aspect_candidate_list_raw.values():
        if len(extracted_term_list) >= frequency_threshold:
            term = extracted_term_list[0]
            qualified_term = QualifiedTerm(term.term, term.pos, len(extracted_term_list),
                                           term_polarity)
            final_aspect_candidates_list.append(qualified_term)
    return final_aspect_candidates_list


def generate_final_opinion_candidates_list(opinion_candidate_list_raw,
                                           final_opinion_candidates_list,
                                           frequency_threshold):
    """Generate final opinion candidates list from raw opinion candidate list.

    Args:
        opinion_candidate_list_raw (dict): key = term, value =
        lists of extracted terms.
        final_opinion_candidates_list (list): list of final opinion candidates
        frequency_threshold (int): minimum freq. for qualifying term
    """
    for extracted_term_list in opinion_candidate_list_raw.values():
        positive_pol = 0
        negative_pol = 0
        if len(extracted_term_list) >= frequency_threshold:
            for ex_term in extracted_term_list:
                curr_polarity = ex_term.term_polarity
                if curr_polarity is not None:
                    if curr_polarity == Polarity.POS:
                        positive_pol = positive_pol + 1
                    if curr_polarity == Polarity.NEG:
                        negative_pol = negative_pol + 1

            # set polarity according majority vote
            term_polarity = Polarity.UNK
            if positive_pol > negative_pol and positive_pol > 0:
                term_polarity = Polarity.POS
            elif negative_pol >= positive_pol and negative_pol > 0:
                term_polarity = Polarity.NEG

            term = extracted_term_list[0]
            qualified_term = QualifiedTerm(term.term, term.pos, len(extracted_term_list),
                                           term_polarity)
            final_opinion_candidates_list.append(qualified_term)
    return final_opinion_candidates_list
